"""

Python Desktop Server - Archive Rendering Tool

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: ArchiveTool.py,v 1.9 2004/04/27 13:31:56 gb Exp $

import re
import sys
import md5
import time
import metakit
import PyDS.Tool
import PyDS.MirrorRegistry

from PyDS.PersistentQueue import PersistentQueue

class ArchiveMirror(PyDS.MirrorRegistry.BaseMirror):

	def _resync(self, tool):
		try:
			archive = PyDS.Tool.getTool('archive')
			archive.suspend()
			res = PyDS.MirrorRegistry.BaseMirror._resync(self, tool)
			for item in self.mirrorstatus.select({'tool':tool.name}):
				if not(archive._hasItem(tool, item.uid)):
					self._addItem(tool, item.uid, tool.getRSSItem(item.uid))
			return res
		finally: archive.resume()
	
	def _fillItem(self, item):
		if item.get('title', '') == '':
			macros = PyDS.Tool.getTool('macros')
			excerpt = macros.unquoteXML(item['description'])
			excerpt = macros.stripHTML(excerpt)
			if len(excerpt) > 80:
				excerpt = excerpt[:75] + ' ...'
			item['title'] = excerpt

	def _signature(self, item):
		self._fillItem(item)
		signature = md5.md5()
		signature.update(item['title'])
		return signature.hexdigest()
	
	def _addItem(self, tool, uniqueId, item):
		self._fillItem(item)
		archive = PyDS.Tool.getTool('archive')
		archive._addItem(tool, uniqueId, item)

	def _updateItem(self, tool, uniqueId, item):
		self._fillItem(item)
		archive = PyDS.Tool.getTool('archive')
		archive._updateItem(tool, uniqueId, item)

	def _deleteItem(self, tool, uniqueId):
		archive = PyDS.Tool.getTool('archive')
		archive._deleteItem(tool, uniqueId)

class ArchiveTool(PyDS.Tool.StandardTool):

	def _createdb(klass, db):
		db.getas("prefs[period:S]")
		db.getas("items[tool:S,uid:S,title:S,link:S,pubtime:I,year:I,quarter:I,month:I,categories[name:S]]")

	_createdb = classmethod(_createdb)

	def _initdb(self):
		try:
			self._acquire()
			self.prefs = self.db.view('prefs')
			if len(self.prefs) == 0:
				self.prefs.append({
					"period":'quarterly',
				})
			self.items = self.db.view('items').ordered(2)
			self._commit()
			self.queue = PersistentQueue(self.name, 'archive')
			self.running = 1
		finally: self._release()
	
	def _initactive(self):
		navigation = self.getToolByNamespace('navigation')
		themes = self.getToolByNamespace('themes')
		navigation._addHook('select', 'archive',
			_('Link to the archive'),
			self._archiveNavHook,
			self.hasArchive)
		themes.hookTemplate(
			'ArchiveRendering',
			_('archive template for single pages'),
			'templates/ArchiveRendering.tmpl'
		)
		themes.hookTemplate(
			'ArchiveIndexRendering',
			_('archive template for archive index'),
			'templates/ArchiveIndexRendering.tmpl'
		)
	
	def _initthread(self, server):
		while server.isRunning():
			time.sleep(10)
			if self.queue.length() and self.running:
				while self.queue.length():
					(cmd, key, data) = self.queue.shift()
					jobf = getattr(self, cmd)
					jobf(key, data)
				self.logVerbose(_('work done, going to sleep again'))
	
	def _status(self):
		html = PyDS.Tool.StandardTool._status(self)
		html += '<ul>'
		html += '<li>There are %d items in the item database</li>' % len(self.items)
		html += self._threadstatus()
		html += '</ul>'
		return html

	# disable background thread, so no rendering actually takes place
	def suspend(self):
		self.running = 0
	
	# ensable background thread, so rendering takes place
	def resume(self):
		self.running = 1
	
	def _rerender(self):
		years = {}
		for item in self.items:
			if not(years.has_key(item.year)):
				years[item.year] = {}
			if not(years[item.year].has_key(item.quarter)):
				years[item.year][item.quarter] = {}
			years[item.year][item.quarter][item.month] = 1
		for year in years.keys():
			for quarter in years[year].keys():
				for month in years[year][quarter].keys():
					self.renderTimeframe(year, quarter, month)
	
	# this method check wether there are entries in the archive
	def hasArchive(self):
		return len(self.items)
	
	def _archiveNavHook(self, qual, value, preset=None):
		if qual == 'html':
			return ''
		elif qual == 'url':
			return self.getStaticUrl()
		else:
			raise ValueError('%s is not a valid qualifier' % qual)

	def _hasItem(self, tool, uniqueId):
		(idx, found) = self.items.locate({
			'tool':tool.name,
			'uid':uniqueId
		})
		return found

	def _addItem(self, tool, uniqueId, item):
		try:
			self._acquire()
			macros = self.getToolByNamespace('macros')
			pubtime = macros.parseTodayFull(item['pubdate'])
			(year, month) = (time.localtime(pubtime))[:2]
			quarter = int((month-1)/3)+1
			categories = map(lambda c: {'name':c}, item['categories'])
			(idx, found) = self.items.locate({
				'tool':tool.name,
				'uid':uniqueId
			})
			if found:
				anitem = self.items[idx]
				anitem.tool = tool.name
				anitem.uid = uniqueId
				anitem.title = macros.unquoteXML(item['title'])
				anitem.link = item['guid']
				anitem.pubtime = pubtime
				anitem.categories = categories
				anitem.year = year
				anitem.quarter = quarter
				anitem.month = month
			else:
				self.items.append({
					'tool':tool.name,
					'uid':uniqueId,
					'title':macros.unquoteXML(item['title']),
					'link':item['guid'],
					'pubtime': pubtime,
					'year': year,
					'quarter': quarter,
					'month': month,
					'categories':metakit.view()
				})
				self._commit()
				(idx, found) = self.items.locate({
					'tool':tool.name,
					'uid':uniqueId
				})
				if found:
					self.items[idx].categories = categories
			self._commit()
			self.renderTimeframe(year, quarter, month)
		finally: self._release()
	
	def _updateItem(self, tool, uniqueId, item):
		self._addItem(tool, uniqueId, item)
	
	def _deleteItem(self, tool, uniqueId):
		try:
			self._acquire()
			(idx, found) = self.items.locate({
				'tool':tool.name,
				'uid':uniqueId
			})
			if found:
				year = self.items[idx].year
				quarter = self.items[idx].quarter
				month = self.items[idx].month
				self.items.delete(idx)
				self._commit()
				self.renderTimeframe(year, quarter, month)
		finally: self._release()

	# This method renders pages overlapping the timeframe in the background
	def renderTimeframe(self, year, quarter, month):
		self.queue.append('_renderTimeframeOnline', (self.getPref('period'), year, quarter, month), None)
		self.queue.append('_renderIndexOnline', 'index', None)
		return self.queue.length()

	def _renderTimeframeOnline(self, key, data):
		(period, year, quarter, month) = key
		req = self._getDummyRequest(datespec=(year, month, 1))
		page = ''
		if period == 'yearly':
			req.setLocalValue('title', _('Archive for %s') % year)
			page = '%s.html' % year
		elif period == 'quarterly':
			req.setLocalValue('title', _('Archive for quarter %s of %s') % (quarter, year))
			page = '%s/%s.html' % (year, quarter)
		elif period == 'monthly':
			req.setLocalValue('title', _('Archive for %s/%s') % (month, year))
			page = '%s/%s.html' % (year, month)
		if page:
			body = req.renderPage('ArchiveRendering',
				period=period,
				year=year,
				quarter=quarter,
				month=month
			)
			html = req.renderPage('BaseRendering',
				body=body,
				category=''
			)
			self.storeHTML(html, page=page)
	
	def _renderIndexOnline(self, key, data):
		req = self._getDummyRequest()
		body = req.renderPage('ArchiveIndexRendering',
			period=self.getPref('period')
		)
		html = req.renderPage('BaseRendering',
			body=body,
			category=''
		)
		self.storeHTML(html)
	
	# this method returns a link tag to a year archive
	def linkYear(self, year, short=0):
		if short:
			return '<a href="%s" title="%s">%s</a>' % (
				self.getStaticUrl('%s.html' % year),
				_('Archive for year %s') % year,
				str(year)
			)
		else:
			return '<a href="%s" title="%s">%s</a>' % (
				self.getStaticUrl('%s.html' % year),
				_('Archive for year %s') % year,
				_('Archive for year %s') % year
			)
	
	# this method returns a link tag to a quarter archive
	def linkQuarter(self, year, quarter, short=0):
		if short:
			return '<a href="%s" title="%s">%s</a>' % (
				self.getStaticUrl('%s/%s.html' % (year, quarter)),
				_('Archive for year %s, quarter %s') % (year, quarter),
				str(quarter)
			)
		else:
			return '<a href="%s" title="%s">%s</a>' % (
				self.getStaticUrl('%s/%s.html' % (year, quarter)),
				_('Archive for year %s, quarter %s') % (year, quarter),
				_('Archive for year %s, quarter %s') % (year, quarter)
			)
	
	# this method returns a link tag to a month archive
	def linkMonth(self, year, month, short=0):
		if short:
			return '<a href="%s" title="%s">%s</a>' % (
				self.getStaticUrl('%s/%s.html' % (year, month)),
				_('Archive for year %s, month %s') % (year, month),
				str(month)
			)
		else:
			return '<a href="%s" title="%s">%s</a>' % (
				self.getStaticUrl('%s/%s.html' % (year, month)),
				_('Archive for year %s, month %s') % (year, month),
				_('Archive for year %s, month %s') % (year, month)
			)
	
	# this method returns all years that have items
	def getYears(self):
		years = {}
		for item in self.items:
			years[item.year] = 1
		liste = years.keys()
		liste.sort(lambda a, b: -1*cmp(a,b))
		return liste
	
	# this method returns all quarters for a given year that have items
	def getQuarters(self, year):
		quarters = {}
		for item in self.items.select({'year':year}):
			quarters[item.quarter] = 1
		liste = quarters.keys()
		liste.sort(lambda a, b: -1*cmp(a,b))
		return liste
	
	# this method returns all months for a given year that have items
	def getMonths(self, year):
		months = {}
		for item in self.items.select({'year':year}):
			months[item.month] = 1
		liste = months.keys()
		liste.sort(lambda a, b: -1*cmp(a,b))
		return liste
	
	# this method returns all items for this period specification
	def getItems(self, period, year, quarter, month):
		ptmin = 0
		ptmax = 0
		items = []
		if period == 'yearly':
			items = self.items.select({'year':year})
		elif period == 'quarterly':
			items = self.items.select({'year':year, 'quarter':quarter})
		elif period == 'monthly':
			items = self.items.select({'year':year, 'month':month})
		liste = []
		for item in items:
			liste.append({
				'tool':item.tool,
				'title':item.title,
				'link':item.link,
				'categories':map(lambda c: str(c.name),item.categories),
				'pubtime':item.pubtime
			})
		liste.sort(lambda a, b: -1*cmp(a['pubtime'], b['pubtime']))
		return liste
	
	# this method returns descriptions of hooked up preferences panels
	# for this tool
	def getPreferencePanels(self):
		return [
			['archive_panel', _('Archives'),
				_('This panel configures how your posting archives look.')]
		]
	
	def archive_panel(self, req):
		macros = self.getToolByNamespace('macros')
		html = '<p>%s' % _('This panel configures your posting archives.')
		html += '<form action="%s" method="POST">' % req.getUrl('archive_redir', tool='archive')
		html += '<table border=0 width="100%">'
		html += '<td>%s</td><td>' % _('period')
		html += '<select name="period">'
		translation = {
			'yearly': _('yearly'),
			'quarterly': _('quarterly'),
			'monthly': _('monthly'),
		}
		for period in ('yearly', 'monthly', 'quarterly'):
			if self.getPref('period') == period:
				html += '<option value="%s" selected>%s' % (period, translation[period])
			else:
				html += '<option value="%s">%s' % (period, translation[period])
		html += '</select>'
		html += '</td></tr><tr><td colspan=2 align="right">'
		html += '<input type=submit name="ArchiveSubmit" value="%s">' % _('save')
		html += '</td></tr>'
		html += '</table>'
		html += '</form>'
		return html
		
	def archive_redir(self, req):
		try:
			self._acquire()
			try:
				self.prefs[0].period = req.getFirstValue('period', 'quarterly')
			except: pass
			self._commit()
			return req.getUrl(
				'panel_html',
				tool='preferences',
				panel='archive_panel',
				fortool='archive'
			)
		finally: self._release()
	
	# this method returns a stored preference value
	def getPref(self, field):
		return getattr(self.prefs[0], field, '')


__desc__ = _("This tool allows to configure the archives of postings in your database.")

PyDS.Tool.registerTool(ArchiveTool, 'archive', _('Archive'), __desc__, 1325)

PyDS.MirrorRegistry.registerMirror('archive', _('This mirror replicates items to the PyDS archive pages'), ArchiveMirror)

