"""
"There's no place like home." -- Dorothy

The DorothyLocker module provides a reimplementation of threading.RLock 
that assists debugging lock contention problems by 

a) checking whether any previous locks have been correctly released 
   before performing subsequent acquire() and release() calls, or 
   whenever sanity_check() is called; and

b) giving stack trace information of the locking call context when 
   another thread has to wait too long for the lock to be released. 

`DorothyRLock` will raise `LockAssertionError` exceptions for any lock 
usage other than:: 

    try: 
        lock.acquire()
        # do stuff
    finally: 
        lock.release()

DorothyRLock does permit `RLock`'s nesting behaviour, without which it'd 
just be a normal `Lock`. Each method must, however, release the lock if 
it acquired it. 

To support environments like PyDS in which an instance's lock is managed 
via bound acquire and release methods, DorothyRLock instances can be 
configured with an ignore list. By ignoring frames from these methods, 
DorothyRLock can concentrate on the frames actually causing the locks and 
releases. For more details, see the implementation of `IgnoreTests`.
"""

import threading
from threading import _Verbose, currentThread, Thread
from thread import allocate_lock
import time
import inspect
import sys
import re
import types

__all__ = ['DorothyRLock', 'LockAssertionError', 'CallingContext']

__revision__ = '$Id: DorothyLocker.py,v 1.8 2004/09/29 11:40:32 gb Exp $'

def ltrim_common(oblist1, oblist2): 
    "Return the arguments with all common first elements removed."
    for pos in range(min(len(oblist1), len(oblist2))): 
        if not oblist1[pos] == oblist2[pos]: 
            break
    else: 
        pos = pos + 1
    return oblist1[pos:], oblist2[pos:]

class FrameInfo(object): 
    "Frozen frame information."

    def __init__(self, frame): 
        "Initialise ourself from a frame."
        object.__init__(self)
        self.frame_id = id(frame)
        self.f_code = frame.f_code
        self.f_lineno = frame.f_lineno

    def __eq__(self, other): 
        "Does this refer to the same frame ID as the other?"
        return self.frame_id == other.frame_id
        
class CallingContext(object): 
    """Object representing the calling context as a tuple of FrameInfo 
    objects, one per calling frame. The context is reversed so that 
    earlier frames are first, not last."""

    def __init__(self, ignores=None): 
        """Distil the calling context, ignoring nominated code objects and 
        function names.
        
        ignores -- a list of function names or code objects to ignore
        """
        object.__init__(self) # good form
        self.when = time.time()
        self.thread = currentThread()
        frame = None
        if ignores is None: 
            ignores = []
        try: 
            stack = inspect.stack(context=0)
            framestack = []
            for frame, filename, lineno, co_name, lines, index in stack: 
                if frame.f_code is self.__init__.im_func.func_code \
                or frame.f_code in ignores or co_name in ignores: 
                    continue
                assert type(frame) == types.FrameType
                framestack.append(FrameInfo(frame))
            framestack.reverse()
            self.framestack = tuple(framestack)
        finally: 
            del frame # apparently helpful with frame objects
    
    def __getslice__(self, i, j): 
        "Get a slice of frames from the context."
        return self.framestack.__getslice__(i, j)

    def __getitem__(self, index): 
        "Get a frame from the context by number (-1 is most recent)."
        return self.framestack[index]

    def __len__(self): 
        "How many frames are in the context?"
        return len(self.framestack)

    def minitrace(self): 
        "Extract a mini-trace from the context."
        tups = []
        for frame in self: 
            tups.append((frame.f_code.co_filename, 
                         frame.f_lineno, 
                         frame.f_code.co_name))
        tups.reverse()
        return tups

    def formatted_minitrace(self): 
        "Return a formatted mini-trace of the context."
        lines = []
        for filename, line, name in self.minitrace(): 
            lines.append('File "%s", line %d, in %s' % (
                  filename, line, name))
        return lines

    def print_minitrace(self, out=None): 
        "Print a mini-trace of the context."
        if out is None: 
            out = sys.stdout
        print >> out, '\n'.join(self.formatted_minitrace())
    
def consolidate_string(string): 
    """Consolidate a string, eliminating redundant whitespace.
    Intended to turn docstrings into single-line strings."""
    return re.sub('\s+', ' ', string).strip()
    
class LockAssertionError(AssertionError): 
    "Raised if DorothyRLock detects a failure to release an acquired lock."
    def __init__(self, lock_context, detection_context, message=""): 
        self.lock_context = lock_context
        self.detection_context = detection_context

        # If the user didn't specify a message, use our doc string.
        if not message: 
            message = consolidate_string(self.__doc__)
        self.message = message

        # Super call; good form, and helps anyone who looks at .args
        AssertionError.__init__(self, lock_context, detection_context, 
                message)

    def formatted_diagnosis(self): 
        "Return lines of a formatted diagnosis of the LockAssertionError."
        name = self.__class__.__name__
        dash = '-' * len(name)
        msg = [dash, name, dash, self.message]
        if self.detection_context: 
            msg.append("Detection context:")
            msg.extend(['  '+line for line in 
                    self.detection_context.formatted_minitrace()])
        if self.lock_context: 
            msg.append("Lock context:")
            msg.extend(['  '+line for line in 
                    self.lock_context.formatted_minitrace()])
        return msg

    def print_diagnosis(self, out=None): 
        "Print a diagnosis of the LockAssertionError."
        if out is None: 
            out = sys.stdout
        print >> out, '\n'.join(self.formatted_diagnosis())

class LeftAcquiringFrame(LockAssertionError): 
    """The lock owner's call stack no longer contains the frame 
    from which the owner acquired the lock. This is almost certainly 
    a failure to release the lock."""

class ReleaseFromSubFrame(LockAssertionError): 
    """A sub-frame of the locking frame tried to release. Either the 
    locking frame failed to release, or the sub-frame failed to 
    acquire."""

class ReAcquireFromSameFrame(LockAssertionError): 
    """A frame tried to acquire again, indicating a probable release 
    in a method which has more than one acquire/release pair or which 
    acquires and releases inside a loop."""

class ReleaseFromForeignThread(LockAssertionError): 
    """An attempt to release the lock from a thread that didn't own it. 
    Strictly speaking, this is probably a failure to acquire."""

class ReleaseOfUnAcquiredLock(LockAssertionError): 
    """An attempt to release an un-acquired lock. Probably caused by 
    failure to acquire."""

class DorothyRLock(_Verbose): 
    """DorothyRLock insists upon release() being called from the same 
    execution frame that acquire()d it in the first place. Frames named 
    in DorothyLocker.IGNORE will be ignored.
    
    If a subframe re-acquires, that's okay, so long as it releases before 
    acquire or release is called via further up the frame stack."""

    def squawk(cls): 
        "Let users know if we're in use."
        att = 'squawked'
        if not hasattr(cls, att): 
            msg = "Enabled verbose lock debugging via DorothyLocker."
            dash = '-' * len(msg)
            print "\n%s\n%s\n%s\n" % (dash, msg, dash)
            setattr(cls, att, True)
    squawk = classmethod(squawk)
        
    def __init__(self, verbose=None, name='(unknown)', ignores=None): 
        "Initialise the DorothyRLock."
        _Verbose.__init__(self, verbose)
        self.name = name
        self.__ignores = [
            self.acquire.func_code, 
            self.release.func_code,
            self.call_context.func_code,
            self.sanity_check.func_code]
        if ignores is not None: 
            self.__ignores.extend(ignores)
        self.__owner = None
        self.__lockstack = []
        self.__block = allocate_lock()
        self.__class__.squawk()
        
    def call_context(self): 
        "Distil our calling context, with instance-specific ignores."
        return CallingContext(ignores=self.__ignores)
        
    def sanity_check(self): 
        """Perform a sanity check: if we're locked by this thread, and 
        the calling frame doesn't share the locking frame, we know the 
        locking frame didn't release.
        
        Be careful calling this if you're about to release."""

        self.__sanity_check('anytime', self.call_context())

    def __sanity_check(self, mode, my_context): 
        """Perform sanity checks when asked by acquire, release, or the 
        public sanity_check method.

        mode -- 'anytime', 'acquire', or 'release'.
        my_thread -- my thread object.
        my_context -- a CallingContext obtained via call_context.
        """
        
        assert mode in ['anytime', 'acquire', 'release']

        owner = self.__owner

        if owner is my_context.thread: 
            lock_context = self.__lockstack[-1]
            # eliminate the common parts of the call stack
            lock_uniques, my_uniques = ltrim_common(lock_context, my_context)
            if lock_uniques:
                # if my_uniques, it's different part of the call stack
                # if not my_uniques, we're in one of our calling frames
                # this is bad nomatter what the sanity check mode
                raise LeftAcquiringFrame, (lock_context, my_context)
            else: 
                if my_uniques: 
                    # we're further down the call stack from the acquire, 
                    # which is: 
                    # * wrong for release, 
                    # * fine for sanity check, and
                    # * fine for acquire
                    if mode == 'release': 
                        raise ReleaseFromSubFrame, (lock_context, my_context)
                else: 
                    # the contexts are identical, which is: 
                    # * perfect for release,
                    # * unusual for sanity_check, and
                    # * highly unusual for acquire
                    if mode == 'acquire':
                        raise ReAcquireFromSameFrame, (lock_context, my_context)
                    
        elif owner is None: 
            # fine for sanity and acquire, bad for release
            if mode == 'release': 
                raise ReleaseOfUnAcquiredLock, (None, my_context)

        else: # foreign owner
            # fine for sanity, acquire will happily block, bad for release
            lock_context = self.__lockstack[-1]
            if mode == 'release': 
                raise ReleaseFromForeignThread, (lock_context, my_context)

    def acquire(self, blocking=1): 
        """Acquire the lock, first checking that any prior locks by this 
        thread have been released if they're not still on our call chain.
        """
        
        # First, get our context. 
        my_context = self.call_context()

        # Are we the owner? 
        # This check is safe because owner is me only if we've called 
        # acquire() from the same thread. 
        if self.__owner is my_context.thread: 
            # Try to figure out whether the previous lock should have been 
            # released. 
            self.__sanity_check('acquire', my_context)

            # If we got this far, we're OK. 
            # Add our details to the lock stack and return success. 
            self.__lockstack.append(my_context)
            if __debug__: 
                self._note("%s.acquire(%s): recursive success", self, blocking)
            return 1
        
        # The lock isn't owned by this thread. The only way to know for
        # sure it isn't owned by anyone else is to acquire it. Let's try 
        # doing it without blocking, first. 
        result = self.__block.acquire(0)
        
        if not result: # Our attempt to lock failed. 
            if not blocking: 
                # If we didn't want a blocking call, we can fail outright here. 
                if __debug__: 
                    self._note("%s.acquire(%s): failure", self, blocking)
                return 0

            # We know for sure someone else had self.__block at the time 
            # we tried to get it, but by now the ownership might have 
            # changed or it might have been released. So, we'll start a 
            # LockWhiner and then perform a blocking wait. 
            whiner = LockWhiner(self, my_context)
            whiner.start()
            result = self.__block.acquire(1)
        
            # We can ONLY get here if we succeeded...
            assert result # ... but you can never be too careful. 
            
            # Stop the whiner. 
            whiner.stop()
        
        # Success! Let's grab the goodies and run. 
        self.__owner = my_context.thread
        self.__lockstack = [my_context]
        if __debug__: 
            self._note("%s.acquire(%s): initial success", self, blocking)
        return 1

    def release(self): 
        """Release the lock, first checking that we're releasing from the 
        same frame that acquired us."""
        
        # First, get our context. 
        my_context = self.call_context()

        # Raise an appropriate exception if we detect a problem or 
        # shouldn't release (say, because we don't own the lock). 
        self.__sanity_check('release', my_context)
        
        # If we got here, all is well. 
        lock_context = self.__lockstack.pop()
        delay = my_context.when - lock_context.when
        if self.__lockstack: 
            if __debug__: 
                self._note("%s.release(): non-final release after %.2fs", 
                    self, delay)
        else: 
            if __debug__: 
                self._note("%s.release(): final release after %.2fs", 
                    self, delay)
            self.__owner = None
            self.__block.release()
            # That better be the last statement that affects state, because
            # a thread waiting in the middle of acquire() will soon be 
            # adjusting the state itself. 

    def __repr__(self): 
        "Return a representation of this object."
        return "<%s owned by %s with count %d>" % (
            self.name, 
            self.__owner and self.__owner.getName(), 
            len(self.__lockstack))

    def lock_details(self): 
        """Return top lock details: owner, lock count, top context, time.
        Because of race conditions, details might not be consistent."""
        try: 
            lock_context = self.__lockstack[-1]
        except IndexError: 
            lock_context = None
        return self.__owner, len(self.__lockstack), lock_context

class LockWhiner(Thread): 
    """Thread that whines until a DorothyRLock.acquire() succeeds, 
    occasionally whining loudly with [hopefully] helpful debugging 
    information about the current owner of the lock."""
    
    def __init__(self, dorothy, whine_context, out=None, every=6, 
            loudly_every=None): 
        """Initialise the LockWhiner.
        
        dorothy -- the lock for which we're waiting
        whine_context -- the call context that's waiting
        out -- file-like-object to whine to
        every -- how many seconds to wait between whines
        loudly_every -- how many seconds to wait between loud whines
        """
        self.dorothy = dorothy
        self.whine_context = whine_context
        self.every = every
        if out is None: 
            out = sys.stdout
        self.out = out
        if loudly_every: 
            self.loudly_every = loudly_every
        else: 
            self.loudly_every = every * 10
        self.active = 1
        self.stop_time = 0 # filled in by stop()
        self.exc_info = None # filled in if necessary by run()

        # Super call -- enforced by Thread.start and other methods
        Thread.__init__(self)
        self.setDaemon(1) # let Python shut down even if we're still active

    def stop(self): 
        """Ask the LockWhiner to stop."""
        self.stop_time = time.time()
        self.active = 0
        delay = self.stop_time - self.whine_context.when
        print >> self.out, "%s: LockWhiner %d stopped for %s; "\
              "acquired by thread %d (%s) after %.2f seconds" % (
              time.ctime(self.stop_time), 
              id(self), 
              self.dorothy.name, 
              id(self.whine_context.thread), 
              self.whine_context.thread.getName(),
              delay)

    def run(self): 
        """Whine until stopped."""
        try: 
            self._run()
        except Exception: 
            # Stash the exception for later if we're supposed to still 
            # be waiting. If not, we probably crashed creating a report 
            # because the lock changed status. 
            if self.active: 
                self.exc_info = sys.exc_info()

    def check(self): 
        """Check for exceptions caught by `run`."""
        if self.exc_info is not None: 
            extype, exvalue, extraceback = self.exc_info
            raise extype, exvalue, extraceback

    def _run(self): 
        """Whine until stopped."""
        out = self.out
        whine_time = self.whine_context.when
        print >> out, "%s: LockWhiner %d started for %s; "\
              "thread %d (%s) waiting" % (\
              time.ctime(whine_time), 
              id(self), 
              self.dorothy.name, 
              id(self.whine_context.thread),
              self.whine_context.thread.getName())
        waits = 0
        louds = int(self.loudly_every/self.every)
        while self.active: 
            time.sleep(self.every)
            if not self.active: 
                break
            waits = waits + 1
            now = time.time()
            print >> out, "%s: LockWhiner %d has been waiting for %.2fs" % (
                  time.ctime(now),
                  id(self), 
                  now - whine_time)
            if not (waits-1) % louds: 
                # Prepare a detailed complaint. This will fail if the 
                # lock is suddenly released and its state changes, 
                # but that won't stop any thread but the whiner. 
                msg = []
                add = msg.append
                add("-"*70)
                add("Just in case this is a deadlock, here's some "\
                    "additional information")
                add("for debugging purposes:\n")
                owner, lock_count, lock_context = \
                       self.dorothy.lock_details()
                if lock_context is None: 
                    time_rep = "(unknown)"
                else: 
                    time_rep = time.ctime(lock_context.when)
                add("Lock most recently obtained at %s "\
                    "by thread id %d, name %s" % (
                        time_rep, id(owner), owner.getName()))
                add("Lock count by this thread: %d" % lock_count)
                if not owner.isAlive(): 
                    add("**THREAD IS DEAD**")
                add("\nLock holding stack:")
                if lock_context is None: 
                    add("(unknown)")
                else:
                    msg.extend(lock_context.formatted_minitrace())
                if waits == 1: 
                    add("\nWaiting thread: id %d, name %s" % (
                        id(self.whine_context.thread), 
                        self.whine_context.thread.getName()))
                    add("\nWaiting stack:")
                    msg.extend(self.whine_context.formatted_minitrace())
                add("-"*70)
                print >> out, '\n'.join(msg)

if __name__ == '__main__': 
    import unittest
    import StringIO

    class CallingContextTests(unittest.TestCase): 
        "Tests of multi-threaded locking behaviour."

        # First, define some methods used by the tests. 
        # Argument funkiness used to pass arguments to CallingContext. 

        def one(self, ignores=None): 
            "Return whatever `two` returns."
            return self.two(ignores)
  
        def two(self, ignores=None): 
            "Return whatever 'three' returns."
            return self.three(ignores)
        
        def three(self, ignores=None):
            "Return `three`'s call context."
            return CallingContext(ignores)

        # Now, the tests: 

        def test_creation(self): 
            "Test call context creation"
            CallingContext()

        def test_basic(self): 
            "Basic context check"
            context = self.one()
            assert context[-1].f_code is self.three.func_code
            assert context[-2].f_code is self.two.func_code
            assert context[-3].f_code is self.one.func_code

        def test_ignorecodes(self): 
            "Make sure ignoring codes works"
            context = self.one(ignores=[self.three.func_code])
            assert context[-1].f_code is self.two.func_code

        def test_ignorenames(self): 
            "Make sure ignoring names works"
            context = self.one(ignores=['three'])
            assert context[-1].f_code is self.two.func_code

        def test_getitem(self): 
            "Get last frame as context[-1]"
            context = self.one()
            context[-1]
            
        def test_len(self): 
            "Return number of frames in context as len(context)"
            context = self.one()
            assert len(context) > 3
            
        def test_minitrace(self): 
            "Context mini-trace"
            context = self.one()
            minitrace = context.minitrace()
            assert isinstance(minitrace, list)
            if hasattr(__builtins__, 'basestring'): 
                stringtype = basestring
            else: 
                stringtype = str
            for filename, lineno, name in minitrace: 
                assert isinstance(filename, stringtype)
                assert isinstance(lineno, int)
                assert isinstance(name, stringtype)
            assert minitrace[0][2] == 'three'
            try: 
                assert minitrace[0][0] == __file__
            except NameError: 
                pass # Python 2.2
            
        def test_formatted_minitrace(self): 
            "Formatted context mini-trace"
            context = self.one()
            formatted_minitrace = context.formatted_minitrace()
            assert isinstance(formatted_minitrace, list)

        def test_print_minitrace(self): 
            "Printed context mini-trace"
            out = StringIO.StringIO()
            context = self.one()
            context.print_minitrace(out)
            assert len(out.getvalue())
            out.close()

    class BasicTests(unittest.TestCase): 
        "Basic tests."

        def test_creation(self): 
            "Creation"
            DorothyRLock()
            
        def test_verbose_creation(self): 
            "Verbose creation"
            DorothyRLock(verbose=True)
            
        def test_named_creation(self): 
            "Named creation"
            DorothyRLock(name='lockName')
            
        def test_verbose_named_creation(self): 
            "Verbose, named creation"
            DorothyRLock(verbose=True, name='lockName')
            
        def test_acquire_release(self): 
            "Basic acquisition and release"
            dorothy = DorothyRLock()
            dorothy.acquire()
            dorothy.release()

        def test_mistaken_release_of_unacquired_lock(self): 
            "React properly to someone releasing an un-acquired lock"

            def mistaken_release_of_unacquired_lock(): 
                "Release a lock without acquiring it."
                dorothy = DorothyRLock()
                dorothy.release()

            self.assertRaises(ReleaseOfUnAcquiredLock, 
                              mistaken_release_of_unacquired_lock)

        def test_formatted_diagnosis(self): 
            "LockAssertionError formatted diagnostics"
            dorothy = DorothyRLock()
            try: 
                dorothy.acquire()
                dorothy.acquire()
            except LockAssertionError, ex: 
                formatted_diagnosis = ex.formatted_diagnosis()
                assert len(formatted_diagnosis)

        def test_printed_diagnosis(self): 
            "LockAssertionError printed diagnostics"
            dorothy = DorothyRLock()
            try: 
                dorothy.acquire()
                dorothy.acquire()
            except LockAssertionError, ex: 
                out = StringIO.StringIO()
                ex.print_diagnosis(out)
                assert len(out.getvalue())
                out.close()
                
    class ReleaseFailureTests(unittest.TestCase):
        "Tests of the release failure detection mechanism."

        def acquire_and_release(self, dorothy): 
            "Acquire and release `dorothy`. Used by tests."
            dorothy.acquire()
            dorothy.release()

        def acquire_with_no_release(self, dorothy): 
            "Acquire, but don't release, `dorothy`. Used by tests."
            dorothy.acquire()
            # didn't call dorothy.release() -- oops!
                
        def release_with_no_acquire(self, dorothy): 
            "Release, but don't acquire, `dorothy`. Used by tests."
            # didn't call dorothy.acquire() -- oops!
            dorothy.release()
            
        def test_boxed_pairs(self): 
            "Permit subframes to re-acquire"
            dorothy = DorothyRLock()
            dorothy.acquire()
            self.acquire_and_release(dorothy)
            dorothy.release()

        def test_release_from_calling_frame(self): 
            "Detect release failure when a calling frame releases"

            def release_from_calling_frame(): 
                "Force `LeftAcquiringFrame`."
                dorothy = DorothyRLock()
                dorothy.acquire()
                self.acquire_with_no_release(dorothy)
                dorothy.release()
            
            self.assertRaises(LeftAcquiringFrame, release_from_calling_frame)

        def test_release_from_sub_frame(self): 
            "Detect acquire failure when a sub-frame releases"

            def release_from_sub_frame(): 
                "Force `ReleaseFromSubFrame`."
                dorothy = DorothyRLock()
                dorothy.acquire()
                self.release_with_no_acquire(dorothy)
                dorothy.release()
            
            self.assertRaises(ReleaseFromSubFrame, release_from_sub_frame)

        def test_reacquire_from_calling_frame(self): 
            "Detect release failure when a calling frame re-acquires"

            def reacquire_from_calling_frame(): 
                "Force `LeftAcquiringFrame`."
                dorothy = DorothyRLock()
                self.acquire_with_no_release(dorothy)
                dorothy.acquire()
            
            self.assertRaises(LeftAcquiringFrame, reacquire_from_calling_frame)
            
        def test_reacquire_from_foreign_frame(self): 
            "Detect release failure when the lock is re-acquired from elsewhere"

            def reacquire_from_foreign_frame():
                "Force `LeftAcquiringFrame`."
                dorothy = DorothyRLock()
                self.acquire_with_no_release(dorothy)
                # the second time around, this is a different *frame* 
                # despite it being the same callable object
                self.acquire_with_no_release(dorothy)

            self.assertRaises(LeftAcquiringFrame, reacquire_from_foreign_frame)
            
        def test_reacquire_from_same_frame(self): 
            "Detect release failure when lock re-acquired from the same frame"

            def reacquire_from_same_frame():
                "Force `ReAcquireFromSameFrame`."
                dorothy = DorothyRLock()
                dorothy.acquire()
                dorothy.acquire()

            self.assertRaises(ReAcquireFromSameFrame, reacquire_from_same_frame)
            
    class LockSittingThread(threading.Thread): 
        "Thread that sits on a lock. Automatically starts itself"

        def __init__(self, lock): 
            "Initialise and start the `LockSittingThread`."
            self.lock = lock
            self.acquired = 0
            self.keepGoing = 1
            threading.Thread.__init__(self)
            self.setDaemon(1) # lets test suite exit if thread still going

            # start, and wait for acquisition
            self.start()
            while not self.acquired: 
                time.sleep(0.1)
            
        def run(self): 
            "In the thread: acquire the lock, wait, and release when told."
            self.lock.acquire()
            self.acquired = 1
            while self.keepGoing: 
                time.sleep(0.1)
            self.lock.release()

        def stop(self): 
            "Ask the thread to stop and release, and wait for it to obey."
            self.keepGoing = 0
            self.join()
        
    class MultiThreadedTests(unittest.TestCase): 
        "Tests of multi-threaded locking behaviour."

        def test_nonblocking_acquire_of_acquired_lock(self): 
            "Fail non-blocking acquires of a lock acquired by another thread"
            dorothy = DorothyRLock()
            sitter = LockSittingThread(dorothy)
            self.assert_(not dorothy.acquire(0))
            sitter.stop()

        def test_foreign_release(self): 
            "Catch one thread releasing a lock acquired by another thread"
            dorothy = DorothyRLock()
            sitter = LockSittingThread(dorothy)
            self.assertRaises(ReleaseFromForeignThread, dorothy.release)
            sitter.stop()

    class IgnoreTests(unittest.TestCase): 
        """Tests of CallContext's ignore behaviour as used by DorothyRLock.
        Also demonstrates the use case for ignores: an instance with a 
        private lock to which it needs to grant controlled access from 
        other objects."""

        def setUp(self): 
            "Set up the private lock."
            self.__dorothy = DorothyRLock(ignores=[
                    self.acquire.func_code, self.release.func_code])

        def acquire(self):
            "Acquire the private lock."
            self.__dorothy.acquire()

        def release(self): 
            "Release the private lock."
            self.__dorothy.release()

        def test_ignores(self): 
            """Verify instance's acquire and release methods are ignored 
            for the purpose of DorothyRLock's context comparisons."""
            self.acquire()
            self.release()

    class LockWhinerTests(unittest.TestCase): 
        "Tests of the `LockWhiner`."

        def test_creation(self): 
            "LockWhiner creation"
            dorothy = DorothyRLock()
            LockWhiner(dorothy, CallingContext())

        def test_fast_creation(self): 
            "Fast LockWhiner creation"
            dorothy = DorothyRLock()
            LockWhiner(dorothy, CallingContext(), every=0.1)

        def test_startstop(self): 
            "LockWhiner start and stop"
            dorothy = DorothyRLock()
            dorothy.acquire()
            # whine on behalf of this context into a StringIO object.
            out = StringIO.StringIO()
            context = CallingContext()
            whiner = LockWhiner(dorothy, context, every=0.1, out=out)
            whiner.start() # start it
            whiner.stop() # ask it to stop
            whiner.join(whiner.every + 1) # wait for it to stop running
            assert not whiner.isAlive() # fail test if we timed out waiting
            whiner.check() # raise any exceptions caught by LockWhiner.run

        def test_whinereports(self): 
            "LockWhiner whining"
            dorothy = DorothyRLock()
            dorothy.acquire()
            out = StringIO.StringIO()
            context = CallingContext()
            whiner = LockWhiner(dorothy, context, every=0.1, out=out)
            whiner.start() # start it

            # Now, make sure there's some output. 
            time.sleep(0.2) # wait a while
            output1 = out.getvalue()
            assert len(output1), "no whining detected"

            time.sleep(0.2) # wait some more
            output2 = out.getvalue()
            assert len(output2)>len(output1), "no additional whining detected"

            # See if we can get a full report
            time.sleep(1)
            output3 = out.getvalue()
            assert output3.find("additional information") >= 0, \
                    "no report detected"

            whiner.stop() # ask it to stop
            whiner.join(whiner.every + 1) # wait for it to stop running
            assert not whiner.isAlive() # fail test if we timed out waiting
            whiner.check() # raise any exceptions caught by LockWhiner.run

    # Run the tests.
    unittest.main()
