"""

Python Desktop Server - Shortcut Editor

Copyright (c) 2002, Garth T Kidd <garth@deadlybloodyserious.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: MeshTool.py,v 1.28 2004/08/04 10:56:36 gb Exp $

import PyDS.Tool
import PyDS.StructuredText
import metakit
import re
import docutils.transforms
import docutils.nodes
import urllib

class TargetVisitor(docutils.nodes.SparseNodeVisitor):
	"""`Visitor` to find targets, as per Georg Bauer's original code."""
	# I'm currently attempting to obsolete this -- Garth

	def __init__(self, document):
		"Initialise the TargetVisitor."
		docutils.nodes.SparseNodeVisitor.__init__(self, document)
		self.targetids = {}
	
	def visit_target(self, node):
		"Visit a target, recording it."
		if node.has_key('name') and node.has_key('id'):
			self.targetids[node['name']] = node['id']
	
	def unknown_visit(self, node):
		pass
	
	def unknown_departure(self, node):
		pass

_toolsWithStrayLinkResolvers = None

def toolsWithStrayLinkResolvers(): 
	"Return a list of tools with old-style stray link resolvers, now deprecated."
	global _toolsWithStrayLinkResolvers 
	# Can someone tell me how to avoid that global? I keep seeing other 
	# code that looks like this, but that doesn't get errors.
	tools = _toolsWithStrayLinkResolvers
	if tools is not None: 
		return tools
	tools = [tool for tool in PyDS.Tool.getTools()
	         if hasattr(tool, 'getCloudUrlMatchingStrayLink')]
	_toolsWithStrayLinkResolvers = tools
	return tools

def orderToolsByContext(foundTools = None, prefer = ''): 
	"Return the prioritised list of tools, with the current tool first."
	if foundTools is None: 
		foundTools = PyDS.Tool.getTools() # ordered by priority
	currentToolName = _flet.current.name
	
	# If one of the found tools matches the current tool, make it first. 
	if [t for t in foundTools if t.name == currentToolName]:
		tools = [_flet.current] 
	else: 
		tools = []
	
	# If one of the found tools matches `prefer`, make it second. 
	if prefer and prefer != currentToolName: 
		prefTools = [t for t in foundTools if t.name == prefer]
		if prefTools: 
			tools.append(prefTools[0])

	# Append the remaining tools and return the list. 
	tools.extend([t for t in foundTools if not t in tools])
	return tools

def scanToolsForSignatures(toolList, *methodList): 
	"""Return a filtered list of tools defining methods. 
	Makes no attempt to tell whether it's a method or an attribute."""
	for method in methodList: 
		toolList = [t for t in toolList if hasattr(t, method)] 
	return toolList

class ToolCache(object): 
	"Cache for tools."
	
	__BorgState = {}
	
	def __new__(cls, *p, **k): 
		"Handle shared state between instances." 
		self = object.__new__(cls, *p, **k)
		self.__dict__ = cls.__BorgState
		return self
	
	def __init__(self): 
		"Initialise the object."
		self.reset()
		
	def reset(self): 
		"Reset the [shared] state."
		self.tools = tools = PyDS.Tool.getTools()
		scan = scanToolsForSignatures
		self.linkTextResolvers = \
			scan(tools, 'getUniqueIdsMatchingLinkText')
		self.unresolvedLinkCreators = \
			scan(tools, 'getDesktopCreateLinkDetailsForProposedLinkText')
		self.cloud404PageRenderer = \
			scan(tools, 'getCloudLinkDetailsFor404')
			
	def contextSortedLinkTextResolvers(self, prefer=''): 
		"Return a context sorted list of link text resolvers."
		return orderToolsByContext(self.linkTextResolvers, prefer)

	def contextSortedUnresolvedLinkCreators(self, prefer=''): 
		"Return a context sorted list of unresolved link creators."
		return orderToolsByContext(self.unresolvedLinkCreators, prefer)
		
	def contextSortedCloud404PageRenderer(self, prefer=''): 
		"Return a context sorted list of renderers for 404 cloud pages."
		return orderToolsByContext(self.cloud404PageRenderer, prefer)
		
# Notes for the new stray link resolution method: 
# OPTIONAL .getCloudLinkDetailsForUniqueId(uniqueId) => (cssclass, URI, title) or (cssclass, None, errmsg)
# cssclass can be None or ''
# OPTIONAL .getDesktopDisplayLinkDetailsForUniqueId(uniqueId) => same
# OPTIONAL .getDesktopCreateLinkDetailsForProposedLinkText(linkText) => same
# Those two MUST raise KeyError if not thrown. 
# REQUIRED .getUniqueIdsMatchingLinkText(linkText) => list of unique IDs
# All the rest of the context (reftime, category, parent) via _flet
# Not all are necessary, eg .getCloudUrlForUniqueId only might be suitable for InterWiki, cached links
# (actually, it'll be necessary because linkText is only fed into getUniqueIds...)

class ObjectDetail(object): 
	"Object representing detail held in a tool database." 
	def __init__(self, tool): 
		"Initialise."
		self._baseTool = tool
		
	def _acquire(self): 
		"Acquire the tool databases."
		return self._baseTool._acquire()
	
	def _release(self): 
		"Release the tool databases."
		return self._baseTool._release()

	def _commit(self): 
		"Commit changes to the tool databases."
		return self._baseTool._commit()
	
class ItemLinksDetail(ObjectDetail): 
	"""Consolidated outbound link details for an object, with __delitem__ 
	etc appropriately mapped back to MeshTool.addLink et al."""
	
	def __init__(self, itemHandle, meshTool): 
		"Initialise the LinkDetail."
		ObjectDetail.__init__(self, meshTool)
		self.itemHandle = itemHandle
		self._links = meshTool.links
		self.refresh()
		
	def _clean(self, linkText): 
		"Clean `linkText`, converting it to a regular string."
		if type(linkText) == type(u''): 
			return str(linkText.encode(_PyDS.documentEncoding))
		return str(linkText)
		
	def refresh(self): 
		"Refresh from the database."
		clean = self._clean
		try: 
			self._acquire()
			outboundLinks = {}
			try: 
				links = self._links.select(originHandle = self.itemHandle)
			except ValueError: 
				links = []
			for l in links: 
				outboundLinks[clean(l.linkText)] = l.destHandle
			self._outboundLinks = outboundLinks
			self._markedLinks = {}
		finally: 
			self._release()

	def mark(self, linkText): 
		"Mark outbound links."
		if not self._outboundLinks.has_key(linkText):
			raise KeyError, linkText
		self._markedLinks[linkText] = 1

	def purge(self): 
		"Delete unmarked outbound links."
		candidates = {}
		for linkText in self._outboundLinks.keys(): 
			candidates[linkText] = 1
		for linkText in self._markedLinks.keys(): 
			del candidates[linkText]
		for linkText in candidates: 
			del self[linkText] # => __delitem__

	def __len__(self): 
		"Returns number of outbound links from this item."
		return len(self._outboundLinks)

	def keys(self): 
		"Returns link texts for outbound links from this item."
		return self._outboundLinks.keys()

	def has_key(self, linkText): 
		"Returns true if this item has an oubound link with text `linkText`."
		linkText = self._clean(linkText)
		return self._outboundLinks.has_key(key)

	def __getitem__(self, linkText): 
		"Returns handle of item pointed to by outbound link `linkText`."
		linkText = self._clean(linkText)
		destHandle = self._outboundLinks[linkText]
		if destHandle == 0: 
			return None
		else: 
			return ItemDetail(handle=destHandle, meshTool=self._baseTool)

	def get(self, linkText, default=None): 
		"Return handle pointed to by `linkText`, with default."
		linkText = self._clean(linkText)
		try: 
			return self.__getitem__(linkText)
		except KeyError: 
			return None
		
	def __fixHandle(self, destHandle): 
		"Find out what the handle behind `destHandle` is."
		if destHandle is None: 
			destHandle = 0
		elif isinstance(destHandle, ItemDetail): 
			destHandle = destHandle.itemHandle
		elif isinstance(destHandle, dict): 
			destHandle = destHandle['handle']
		return destHandle
	
	def __setitem__(self, linkText, destHandle): 
		"Asserts that this item refers to `destHandle` as `linkText`."
		destHandle = self.__fixHandle(destHandle)
		linkText = self._clean(linkText)
		currentDestinationHandle = self._outboundLinks.get(linkText)
		if currentDestinationHandle == destHandle: 
			return
		try: 
			self._acquire()
			self._outboundLinks[linkText] = destHandle
			if currentDestinationHandle is not None: 
				# need to update old record
				spec = {
					'originHandle': self.itemHandle, 
					'destHandle': currentDestinationHandle,
					'linkText': linkText
					}
				(idx, found) = self._links.locate(spec)
				if found: 
					if found>1: 
						raise AssertionError, ("Too many matches", spec)
					self._links[idx].destHandle = destHandle
					self._commit()
			else: 
				self._links.append(originHandle = self.itemHandle, 
				                  linkText = linkText, 
				                  destHandle = destHandle)
				self._commit()
		finally: 
			self._release()

	def __delitem__(self, linkText): 
		"Assert that the item doesn't refer outbound by `linkText`."
		linkText = self._clean(linkText)
		if not self._outboundLinks.has_key(linkText): 
			raise KeyError, linkText
		del self._outboundLinks[linkText]
		try: 
			self._acquire()
			spec = {
					'originHandle': self.itemHandle, 
					'linkText': linkText
					}
			# Constructed to wipe out as many as are defined; less brittle 
			# than what I've done in __setitem__, but I can't be that nasty
			# ALL the time.
			while 1: 
				(idx, found) = self._links.locate(spec)
				if not found: 
					break
				self._links.delete(idx)
			self._commit()	
		finally: 
			self._release()

	def setdefault(self, linkText, destHandle=None): 
		"Set destination of `linkText` to `destHandle` if it doesn't exist."
		destHandle = self.__fixHandle(destHandle)
		if type(linkText) == type(u''):
			linkText = linkText.encode(_PyDS.documentEncoding)
		linkText = str(linkText)
		try: 
			self._acquire()
			match = self._links.select(originHandle = self.itemHandle, linkText=linkText)
			if not match: 
				self._links.append(originHandle = self.itemHandle, 
							  linkText = linkText, 
							  destHandle = destHandle)
				self._outboundLinks[linkText] = destHandle
				handle = destHandle
			else: 
				assert len(match)==1
				handle = match[0].destHandle
			if handle==0: 
				return None
			return ItemDetail(handle=handle, meshTool=self._baseTool)
		finally: 
			self._release()

class ItemDetail(ObjectDetail): 
	"MeshTool's view of another tool's item."
	
	def __init__(self, homeTool=None, uniqueId=None, meshTool=None, handle=None): 
		"Initialises the Item. Creates a database entry if necessary."
		if meshTool is None: 
			meshTool = PyDS.Tool.getTool('mesh')
		ObjectDetail.__init__(self, meshTool) # also sets self.baseTool
		assert handle is not None or (homeTool is not None and uniqueId is not None)
		if handle is not None: 
			assert not isinstance(handle, ItemDetail) 
			#print "ItemDetail.__init__(handle=%s)" % repr(handle)
			itemDetails = meshTool.getItemDetailsByHandle(handle)
			homeTool = itemDetails['homeTool']
			uniqueId = itemDetails['uniqueId']
		self.items = meshTool.items
		self.homeTool = homeTool
		self.uniqueId = uniqueId
		try: 
			self._acquire()
			spec = {'homeTool': homeTool, 'uniqueId': uniqueId}
			idx = self.items.find(spec)
			if idx < 0: 
				itemHandle = self.itemHandle = \
					meshTool.newItemHandle(acquireAndCommit=0)
				ret = self.items.append({
					'handle': itemHandle,
					'homeTool': homeTool, 
					'uniqueId': uniqueId, 
					'exists': 1}) 
				self._commit()

			else: 
				itemHandle = self.itemHandle = self.items[idx].handle
		finally: 
			self._release()
			
		self.links = ItemLinksDetail(itemHandle, meshTool)
		self.uri = meshTool.uriForHandle(itemHandle)

	def getCloudLinkDetails(self): 
		"Get the cloud link details for this item."
		tool = PyDS.Tool.getTool(self.homeTool)
		return tool.getCloudLinkDetailsForUniqueId(self.uniqueId)

	def getDesktopDisplayLinkDetails(self):
		"Get the display/edit link details for this item."
		tool = PyDS.Tool.getTool(self.homeTool)
		return tool.getDesktopDisplayLinkDetailsForUniqueId(self.uniqueId)

class StrayLinkResolver(docutils.nodes.SparseNodeVisitor):
	"""`Visitor` to resolve stray links.
	
	Intended to facilitate links to other weblog postings and anything
	else another tool might come up with. Say, Wiki postings."""

	def __init__(self, document, targetVisitor, itemDetail, meshTool):
		"""Initialise the `StrayLinkResolver` with a given `TargetVisitor`
		instance. The latter should have already been walked over the
		document. `itemDetail` should be an `ItemDetail` object corresponding 
		to the source being rendered, or None."""
		# PyDS.Tool.traceargs()
		docutils.nodes.SparseNodeVisitor.__init__(self, document)
		self.targetVisitor = targetVisitor
		self.foundTools = toolsWithStrayLinkResolvers()
		self.meshTool = meshTool
		self.itemDetail = itemDetail
		self.itemLinks = itemDetail.links
		toolCache = ToolCache()
		self.linkTextResolvers = toolCache.contextSortedLinkTextResolvers()
		prefs = meshTool.getPreferencesTool()
		targ404 = prefs.get('mesh', 'targ404')
		targcreate = prefs.get('mesh', 'targcreate')
		self.unresolvedLinkCreators = toolCache.contextSortedUnresolvedLinkCreators(targ404)
		self.cloud404PageRenderer = toolCache.contextSortedCloud404PageRenderer(targcreate)
	
	# Log methods
	def log(self, f, *a): self.meshTool.log(f, *a)
	def logError(self, f, *a): self.meshTool.logError(f, *a)
	def logVerbose(self, f, *a): self.meshTool.logVerbose(f, *a)
		
	def purge(self): 
		"""Purge unmarked outbound links from the item for which we 
		were initialised."""
		self.itemLinks.purge()

	def resolveStrayLink(self, stray): 
		"""Resolve a stray by consulting all tools capable of doing so."""
		# First, try the cache...
		prevDest = self.itemLinks.setdefault(stray, 0)
		# side-effect of the above: sets dest to 0 if not found
		self.itemLinks.mark(stray)

		if prevDest is not None: 
			try: 
				if _flet.desktop:
					return prevDest.getDesktopDisplayLinkDetails()
				else: 
					return prevDest.getCloudLinkDetails()
			except (KeyError, ValueError): 
				self.logError(_("Cached link destination %s:%s no longer exists."),
				              prevDest.homeTool, prevDest.uniqueId)
				self.meshTool.delItem(prevDest)

		# Gather list of potential matches. 
  		matches = []
		for tool in self.linkTextResolvers: 
			thisMatches = tool.getUniqueIdsMatchingLinkText(stray)
			for match in thisMatches: 
				matches.append((tool, match))

		# If we have no matches, return something appropriate. 
		if len(matches) == 0: 
			# Select which tools to ask, and how to ask them.
			if _flet.desktop: 
				def getDetails(tool, stray): 
					return tool.getDesktopCreateLinkDetailsForProposedLinkText(stray)
				tools = self.unresolvedLinkCreators
			else: 
				def getDetails(tool, stray): 
					return tool.getCloudLinkDetailsFor404()
				tools = self.cloud404PageRenderer

			# Iterate through the tools, trying to get link 
			# details from the named method. 
			for tool in tools: 
				try: 
					return getDetails(tool, stray) 
				except (KeyError, ValueError): 
					pass

			# Cop out if we couldn't find anything. 
			return ('undefinednode', 
			        'about:blank', 
				_('PyDS failed to figure out somewhere better to send you. Sorry about that.'))

		# If we have matches, return something appropriate.
		# TODO: this is not a good way of handling ambiguity. 
		ambiguous = len(matches) > 1
		homeTool, uniqueId = matches[0]
		destHandle = self.meshTool.addItem(homeTool, uniqueId) # won't duplicate
		self.itemLinks[stray] = destHandle
		if _flet.desktop: 
			linkDetails = homeTool.getDesktopDisplayLinkDetailsForUniqueId(uniqueId)
		else: 
			linkDetails = homeTool.getCloudLinkDetailsForUniqueId(uniqueId)
		return linkDetails

	def visit_reference(self, node):
		"""Visit references, registering links and fixing unresolvable references."""
		refname = node.get('refname') # internal references
		refuri = node.get('refuri') # external references

		def applymatch(node, match): 
			if len(match) == 3: 
				cssclass, uri, title = match
				node['title'] = title
			else: 
				cssclass, uri = match
			node.resolved = 1
			node['class'] = cssclass
			node['refuri'] = uri
			try: 
				del node['refname']
			except KeyError: 
				pass
			
		# If the target begins with ``pyds:/``, we want to intercept it. 
		if refuri is not None and refuri.startswith("pyds:/"): 
			match = ('undefinednode', refuri, 'internal link not resolvable')
			try: 
				itemDetails = self.meshTool.getItemDetailsByUri(str(refuri))
				homeTool = PyDS.Tool.getTool(itemDetails['homeTool'])
				uniqueId = itemDetails['uniqueId']
				if _flet.desktop: 
					match = homeTool.getDesktopDisplayLinkDetailsForUniqueId(uniqueId)
				else: 
					match = homeTool.getCloudLinkDetailsForUniqueId(uniqueId)
			except KeyError: 
				# TODO: What should we do here? Really? 
				# Currently: leave it with default match defined above
				pass

			applymatch(node, match)
			return

		# If it has been apparently resolved, return. In testing,
		# this has been seen true with refname and refuri both 
		# undefined. I don't yet know why. 
		if node.resolved: 
			return

		# If it's internal and we can find something resembling a 
		# target, leave it alone. This matches behaviour in 
		# docutils.transforms.universal.FinalChecksVisitor. 
		if refname is not None: 
			if self.document.nameids.get(refname) is not None: 
				return

		# If it's resolved, but to something not resembling a URI, assume 
		# it's really intended to be internal. Note that there's no error 
		# checking, here; if we don't find a match, you'll end up with the 
		# docutils HTML writer default of pointing to a relative document 
		# -- which probably won't exist. 
		if refuri is not None:
			if node.resolved:
				(method, rest) = urllib.splittype(refuri)
				if method is None: 
					match = self.resolveStrayLink(refuri)
					if match: 
						applymatch(node, match)
						#cssclass, uri = match
						#node['class'] = cssclass
						#node['refuri'] = uri
			return # whether resolved or not
		
		if node.hasattr('refid'): 
			# This is some kind of internal link, so far seen only in tables of 
			# content. 
			return
		
		# THIS SHOULD BE OBSOLETE: 
		# If it's resolved, or doesn't target a name, or if we found an 
		# explicit target, return. 
		if node.resolved \
			or not node.hasattr('refname') \
			or self.targetVisitor.targetids.has_key(node['refname']):
			# Let me get this right. If there's no refname, it's 
			# external. So, this is: if it's resolved (implying 
			# external), or it's external, or if we found the 
			# target with targetids, then we object. 
			basic = _("Assertion failure in stray link resolver")
			if refname and self.targetVisitor.targetids.has_key(node['refname']):
				self.meshTool.log("%s: refname %s in targetids but not document.nameids",
				    basic, repr(refname))
			else: 
				self.meshTool.log("%s: refname=%s, refuri=%s, node.resolved=%s, "\
						"targetids.get(%s)=%s, text=%s, itemDetail=%s:%s", 
					basic, repr(refname), repr(refuri), repr(node.resolved),
					repr(refname), 
					repr(self.targetVisitor.targetids.get(refname)),
					repr(node.__dict__),
					self.itemDetail.homeTool, self.itemDetail.uniqueId)
			return

		# Only here do we try and resolve things internally. 
		match = self.resolveStrayLink(node['refname'])
		if match:
			applymatch(node, match)

		# Otherwise, just leave it alone.
			
	def unknown_visit(self, node):
		pass
	
	def unknown_departure(self, node):
		pass

class StrayLinkTransform(docutils.transforms.Transform):
	"""`Transform` to resolve stray links across a document."""

	# should run after all other transformations, but before the
	# universal.Final stuff
	default_priority = 800

	def apply(self):
		if toolsWithStrayLinkResolvers(): 
			# Where's the mesh tool? 
			try: 
				meshTool = StrayLinkResolver.meshTool
			except AttributeError: 
				meshTool = StrayLinkResolver.meshTool = PyDS.Tool.getTool('mesh')

			# Are we registering links? 
			itemDetail = itemKeys = None
			try: 
				itemKeys = ( _flet.homeTool.name, _flet.uniqueId, meshTool )
			except KeyError:
				pass

			if itemKeys is not None: 
				itemDetail = ItemDetail(*itemKeys)
					
			# Modify the document... 
			targetVisitor = TargetVisitor(self.document)
			self.document.walk(targetVisitor)
			slr = StrayLinkResolver(self.document, targetVisitor, itemDetail, meshTool)
			self.document.walk(slr)

			# ... and purge no-longer-used outbound link entries. 
			slr.purge()

class MeshTool(PyDS.Tool.StandardTool):

	def _createdb(klass, db):
		db.getas("items[homeTool:S,uniqueId:S,handle:I,exists:I]")
		db.getas("links[originHandle:I,linkText:S,destHandle:I]")
		db.getas("state[maxhandle:I]")

	_createdb = classmethod(_createdb)

	def _initdb(self):
		"Initialise MeshTool databases."
		try:
			self._acquire()
			self.items = self.db.view('items').ordered(2)
			self.links = self.db.view('links').ordered(1)
			self.state = self.db.view('state')
			if len(self.state) == 0: 
				self.state.append(maxhandle=0)
				self._commit()
			# Texts for which we lack homeTool and uniqueId can't depend on the 
			# MeshTool's technology for being intelligently re-rendered. 
			# Known texts are registered in items. 
			# The links table gathers the links. If destHandle is 0, the link 
			# is not resolved. 
			self.importantTool = 1
		finally: self._release()
	
	def _dumpcontents(self): 
		"Handy debug code for xmlrpc."
		try: 
			self._acquire()
			items = [self._getItemDetailsFromRow(i) for i in self.items]
			links = [{'originHandle':l.originHandle, 
			          'linkText':l.linkText,
					  'destHandle':l.destHandle} for l in self.links]
			return {'items': items, 'links': links}
		finally: self._release()
		
	def _clearcontents(self): 
		"Clear the content."
		try: 
			self._acquire()
			for i in range(len(self.links)): 
				self.links.delete(0)
			for i in range(len(self.items)): 
				self.items.delete(0)
			return 1
		finally: self._release()

	def _backup(self):
		"Return backup state for MeshTool tables."
		l = self._backupTable('items')
		l.extend(self._backupTable('links'))
		l.extend(self._backupTable('state'))
		return l
		
	def _status(self):
		"Return status for the status tool."
		def p(s): 
			return '<p>' + s + '</p>'
		def li(s): 
			return '<li>' + s + '</li>'
		try:
			self._acquire()
			html = PyDS.Tool.StandardTool._status(self)
			html += '<ul>'
			html += li(_('There are %d tools registered for stray link resolving.') % len(toolsWithStrayLinkResolvers()))
			html += li(_('There are %d known texts.') % len(self.items))
			html += li(_('There are %d known links.') % len(self.links))
			html += '</ul>'
			return html
		finally: self._release()

	def newItemHandle(self, acquireAndCommit=1): 
		"Return a new item handle."
		try: 
			self._acquire()
			return self.__newItemHandle()
		finally: 
			self._commit()
			self._release()

	def __newItemHandle(self): 
		"Return a new item handle, without acquire/commit."
		handle = self.state[0].maxhandle + 1
		self.state[0].maxhandle = handle
		return handle
		
	def getItemDetailsByUri(self, uri): 
		"""Fetch item details by pseudo URI.
		
		Raises ValueError if it can't be parsed,
		KeyError if a handle for the parsed toolName and uniqueId can't be found."""
		
		e = r'pyds\:(?://(?P<hostName>[^/]+))?/(?P<toolName>[^/]+)/(?P<uniqueId>.*)'
		m = re.match(e, uri)
		if m is None: 
			raise ValueError, uri
		g = m.groupdict()
		return self.getItemDetailsByToolAndId(g['toolName'], g['uniqueId'])
		
	def handleForUri(self, uri): 
		"Fetch handle by pseudo URI"
		return self.getItemDetailsByUri(uri)['handle']
		
	def uriForHandle(self, handle): 
		"Return a pseudo-URI for the handle."
		try: 
			self._acquire()
			return self.__uriForHandle(handle)
		finally:
			self._release()

	def __uriForHandle(self, handle): 
		"Like `uriForHandle`, but doesn't acquire and release."
		idx = self.items.find({'handle': handle})
		if idx<0: 
			return ''
		i = self.items[idx]
		return self.__uriForHomeToolAndUniqueId(i.homeTool, i.uniqueId)

	def __uriForHomeToolAndUniqueId(self, homeTool, uniqueId): 
		return 'pyds:/%s/%s' % (homeTool, uniqueId)

	def addItem(self, homeTool, uniqueId, exists=1): 
		"Add an item."
		try: 
			self._acquire()
			return self.__addItem(homeTool, uniqueId, exists)
		finally: self._release()
	
	def __addItem(self, homeTool, uniqueId, exists=1): 
		"Add an item, committing but without acquire/release."
		if isinstance(homeTool, PyDS.Tool.StandardTool):
			homeTool = homeTool.name
		if type(uniqueId) != type(''):
			uniqueId = str(uniqueId)
		idx = self.items.find({
			'homeTool': homeTool, 
			'uniqueId': uniqueId})
		if idx < 0: 
			handle = self.__newItemHandle()
			ret = self.items.append({
				'handle': handle,
				'homeTool': homeTool, 
				'uniqueId': uniqueId, 
				'exists': exists})
			self._commit()
			return handle
		else: 
			return self.items[idx].handle

	def __decodedHandle(self, handleOrUniqueId, homeTool): 
		"Return a handle, for methods with handleOrUniqueId, homeTool=None."
		if isinstance(handleOrUniqueId, ItemDetail): 
			assert not homeTool
			return handleOrUniqueId.itemHandle
		if isinstance(handleOrUniqueId, int): 
			assert not homeTool
			return handleOrUniqueId
		else: 
			assert homeTool
			if isinstance(homeTool, PyDS.Tool.StandardTool):
				homeTool = homeTool.name
			return self.getItemDetailsByToolAndId(homeTool, handleOrUniqueId)['handle']

	def renItem(self, uniqueId, homeTool, newUniqueId): 
		"""Change an item's `uniqueId`."""
		try: 
			self._acquire()
			idx = self.items.find({
				'homeTool': homeTool, 
				'uniqueId': uniqueId})
			if idx < 0: 
				raise KeyError, (homeTool, uniqueId)
			self.items[idx].uniqueId = newUniqueId
			self._commit()
			return newUniqueId
		finally: self._release()
		
	def __renItem(self, uniqueId, homeTool, newUniqueId): 
		"""`renItem` without acquire/release"""
		
	def delItem(self, handleOrUniqueId, homeTool=None, abortIfTarget=0): 
		"""Delete an item, unless inbounds and `abortIfTarget`.
		
		Returns: 
			0	 = didn't find the handle
			1    = deleted
			list = map of inbounds as per `getInboundLinks`."""
		try: 
			self._acquire()
			inbounds = self.__delItem(handleOrUniqueId, homeTool, abortIfTarget=0)
		finally:
			self._release()

	def __delItem(self, handleOrUniqueId, homeTool=None, abortIfTarget=0): 
		"""`delItem` without acquire/release"""
		handle = self.__decodedHandle(handleOrUniqueId, homeTool)
		links, items = self.links, self.items
		inboundHandles = self.__getInboundLinks(handle)
		if abortIfTarget and inboundHandles: 
			pass
		else: 
			# Re-target all inbound links to 0
			while 1: 
				idx = links.find({'destHandle': handle})
				if idx < 1: 
					break
				links[idx].destHandle = 0
			# Find the item entry, and delete it
			idx = items.find({'handle': handle})
			if idx < 0: 
				self.logVerbose(_("delItem called for non-existent handle %d!?"), handle)
			else: 
				items.delete(idx)
				self._commit()
			# Log, and re-render any items that had linked to the deleted item
			if inboundHandles: 
				self.logVerbose(_("Item deleted. Re-rendering %d item(s) that had linked to it."), 
					 len(inboundHandles))
				self.meshTool.renderItems(inboundHandles)
			else: 
				self.logVerbose(_("Item deleted. No re-rendering required."))
		return inboundHandles
		
	def _getItemDetailsFromRow(self, row): 
		"Fetch item details out of a row from the items table."
		details = {}
		for key in ['homeTool', 'uniqueId', 'exists', 'handle']: 
			details[key] = getattr(row, key)
		details['uri'] = self.uriForHandle(row.handle)
		return details
		
	def getItemDetailsByToolAndId(self, homeTool=None, uniqueId=None):
		"""Get item details by tool and id, defaulting to current tool and id
		as [hopefully] supplied to `StructuredText.renderText`."""
		try: 
			self._acquire()
			return self.__getItemDetailsByToolAndId(homeTool, uniqueId)
		finally: self._release()

	def __getItemDetailsByToolAndId(self, homeTool=None, uniqueId=None):
		"`getItemDetailsByToolAndId` without acquire/release"
		if homeTool is None: 
			homeTool = _flet.homeTool
		if isinstance(homeTool, PyDS.Tool.StandardTool):
			homeTool = homeTool.name
		if uniqueId is None: 
			uniqueId = _flet.uniqueId
		idx = self.items.find({
			'homeTool': homeTool, 
			'uniqueId': uniqueId})
		if idx < 0: 
			raise KeyError, (homeTool, uniqueId)
		return self._getItemDetailsFromRow(self.items[idx])

	getItem = getItemDetailsByToolAndId
		
	def getItemDetailsByHandle(self, handle): 
		"Get item details by their MeshTool handle."
		try: 
			self._acquire()
			return self.__getItemDetailsByHandle(handle)
		finally: self._release()

	def getItemDetailsByHandles(self, handles): 
		"sequence version of `getItemDetailsByHandle`."
		try: 
			self._acquire()
			return [self.__getItemDetailsByHandle(handle) 
			        for handle in handles]
		finally: self._release()
		
	def __getItemDetailsByHandle(self, handle): 
		"`getItemDetailsByHandle` without acquiring/releasing the database"
		idx = self.items.find({'handle':handle})
		if idx<1: 
			raise KeyError, handle
		return self._getItemDetailsFromRow(self.items[idx])

	def getOutboundLinksByHandle(self, handle): 
		"Get a dictionary mapping outbound link text to destination handle."
		try: 
			self._acquire()
			outboundLinks = {}
			for l in self.links.select(originHandle = handle): 
				outboundLinks[l.linkText] = l.destHandle
			return outboundLinks
		finally: self._release()

	def getInboundLinks(self, handleOrUniqueId, homeTool=None): 
		"Get a dictionary mapping inbound link text to origin handle."
		try: 
			self._acquire()
			return self.__getInboundLinks(handleOrUniqueId, homeTool)
		finally: self._release()

	def __getInboundLinks(self, handleOrUniqueId, homeTool=None): 
		"`getInboundLinks` without acquire/release"
		handle = self.__decodedHandle(handleOrUniqueId, homeTool)
		inboundLinks = {}
		for l in self.links.select(destHandle = handle): 
			originMap = inboundLinks.setdefault(l.originHandle, [])
			originMap.append(l.linkText)
		return inboundLinks
	
	def clear_redir(self, req):
		self._clearcontents()
		return req.getUrl()
		
	def index_html(self, req):
		req.setLocalValue('title', _('Mesh Tool'))
		html = []
		a = html.append
		p = lambda t: html.append('<p>%s</p>' % (t,))
		a('<h2>%s</h2>' % _('Status'))
		a(self._status())
		signatureSpecs = [
			('getCloudUrlMatchingStrayLink', _('old stray link resolver')),
			('getUniqueIdsMatchingLinkText', _('resolve stray link')),
			('getCloudLinkDetailsForUniqueId', _('return cloud link')),
			('getCloudLinkDetailsFor404', _('cloud 404 page renderer')), 
			('getDesktopDisplayLinkDetailsForUniqueId', _('return show/edit link')),
			('getDesktopCreateLinkDetailsForProposedLinkText', _('return create link')),
			('renderItemByUniqueId', _('render item'))
			]
		signatureRadios = {
			'getDesktopCreateLinkDetailsForProposedLinkText': 'targcreate',
			'getCloudLinkDetailsFor404': 'targ404'
			}
		signatures = [signatureSpec[0] for signatureSpec in signatureSpecs]
		signatureDescriptions = dict(signatureSpecs)
		tools = PyDS.Tool.getTools()
		toolHasSig = {}
		sigToolBuckets = {}
		for signature in signatures: 
			sigToolBuckets[signature] = {}
		for tool in tools: 
			toolHasSig[tool] = False
			for signature in signatures: 
				if hasattr(tool, signature): 
					sigToolBuckets[signature][tool] = True
					toolHasSig[tool] = True
		toolsWithSigs = [tool for tool in tools if toolHasSig[tool]]
					
		a('<h2>%s</h2>' % _('Tools with stray link resolution methods:'))
		a('<table>')
		def row(rowcells): 
			a('<tr>')
			a('\n'.join(['  %s' % (cell,) for cell in rowcells]))
			a('</tr>')
		def cell(content, isHeader=False, **atts): 
			tag = 'td'
			if isHeader: tag = 'th'
			attdefs = [' %s=%s' % (attn, attd) for (attn, attd) in atts.items()]
			attdef = ''.join(attdefs)
			return '<%s%s>%s</%s>' % (tag, attdef, content, tag)
		headers = [cell(_('signature'), isHeader=True, align='left')]
		headers.extend([cell(tool.name, isHeader=True, align='left') 
		                for tool in toolsWithSigs])
		row(headers)
		for signature in signatures: 
			sigToolBucket = sigToolBuckets[signature]
			cells = []
			ca = lambda content, **atts: cells.append(cell(content, **atts))
			signatureDescription = signatureDescriptions[signature]
			ca('<span title="%s">%s</span>' % (
				signature, 
				signatureDescription),
				isHeader=True,
				align='left')
			for tool in toolsWithSigs: 
				if sigToolBucket.get(tool): 
					disp = 'yes'
				else: 
					disp = 'no'
				if signature in signatureRadios.keys() and disp=='yes': 
					pref = signatureRadios[signature]
					prefs = self.getPreferencesTool()
					target = prefs.get('mesh', pref)
					if target == tool.name: 
						disp = '<strong>%s</strong>' % disp
					disp = '<a href="%s" class="%s">%s</a>' % (
						req.getUrl('setdefaulttool_redir', 
						           pref=pref, 
							   target=tool.name), 
						'meshsetdefaulttool', 
						disp)
				ca(disp)
			row(cells)
		a('</table>')
		p('<a href="%s">%s</a>' % (
			req.getUrl('clear_redir'),
			_('Click here to clear all cached links')))
		return req.renderPage('BaseTemplate', body='\n'.join(html))

	def setdefaulttool_redir(self, req): 
		pref = req.getFirstValue('pref')
		target = req.getFirstValue('target')
		prefs = self.getPreferencesTool()
		prefs.set('mesh', pref, target)
		return req.getUrl('index_html')

	def items_html(self, req): 
		req.setLocalValue('title', _('Mesh Tool'))
		html = ['<h2>Items</h2>']
		html.append('<p>%s</p>' % \
			_("Items that no longer exist will be cleaned " \
			  "and </strike>struck through</strike>."))
		html.append('<table>')
		html.append('<tr><th>handle</th><th>tool</th><th>uniqueId</th></tr>')
		self._acquire()
		methods = {}
		for item in self.items: 
			link = uniqueId = item.uniqueId
			homeTool = item.homeTool
			cells = ['<a href="%s">%s</a>' % (
			             self.getUrl('inbound_html', handle=item.handle), 
			             item.handle), 
			         homeTool]
			try: 
				method = methods[item.homeTool]
			except KeyError: 
				try: 
					tool = self.getToolByNamespace(homeTool)
					method = tool.getDesktopDisplayLinkDetailsForUniqueId
				except (AttributeError, KeyError): 
					method = None
				if method is not None: 
					try: 
						cssclass, target, title = method(uniqueId)
						link = '<a class="%s" title="%s" href="%s">%s</a>' % (
							cssclass, 
							title, 
							target, 
							uniqueId)
					except KeyError: 
						self.__delItem(item.handle)
						link = '<strike>%s</strike>' % uniqueId
					cells.append(link)
			html.append('<tr>' +  \
			            ''.join(["<td>%s</td>" % cell for cell in cells]) + \
						'</tr>')
		self._release()
		html.append('</table>')
		return req.renderPage('BaseTemplate', body='\n'.join(html))

	def inbound_html(self, req): 
		macros = self.getToolByNamespace('macros')
		req.setLocalValue('title', _('Mesh Tool'))
		handle = int(req.getFirstValue('handle'))
		item = self.getItemDetailsByHandle(handle)
		inbound = self.getInboundLinks(handle)
		req.setLocalValue('title', _('Mesh Tool'))
 		html = []
		a = html.append
		def p(s): 
			a('<p>%s</p' % s)
		p(_("Inbound links for %s") % (item['uri']))
		import pprint
		a('<pre>' + macros.quoteHTML(pprint.pformat(inbound)) + '</pre>')
		return req.renderPage('BaseTemplate', body='\n'.join(html))
		
	def renderItemsWithUnresolvedLinks(self):
		"Re-render items with unresolved links."
		try: 
			self._acquire()
			handles = dict([(l.originHandle, 1) 
			                for l in self.links
			                if l.destHandle == 0]).keys()
		finally: 
			self._release()
		self.logVerbose("Re-rendering %d item(s) with unresolved links", len(handles))
		return [handles, self.renderItems(handles)]

	def renderInbound(self, handleOrUniqueId, homeTool=None): 
		"Re-render items pointing to this item."
		handle = self.__decodedHandle(handleOrUniqueId, homeTool)
		inbound = self.getInboundLinks(handle)
		self.logVerbose("Re-rendering %d item(s) referring to handle %d", 
		                len(inbound), handle)
		if inbound: 
			self.renderItems(inbound)
			
	def renderItems(self, handles=None): 
		"Render items associated with the `handles`, all by default."
		try: 
			self._acquire()
			methods = {}
			rendered = []
			if handles is None: 
				handlemap = [(self.items[i].handle, i) 
				             for i in range(len(self.items))]
			else: 
				handlemap = [(handle, self.items.find({'handle': handle})) 
				             for handle in handles]
			self.logVerbose(_("Re-rendering %d item(s)."), len(handlemap))
			for handle, index in handlemap: 
				item = self.items[index]
				toolName = item.homeTool
				try: 
					method = methods[toolName] 
				except KeyError: 
					tool = self.getToolByNamespace(toolName)
					try: 
						method = tool.renderItemByUniqueId
					except KeyError: 
						method = None
					methods[toolName] = method
				if method is not None: 
					print "mesh.renderItems: requested re-render for", 
					method(item.uniqueId)
					rendered.append(handle)
				else: 
					print "mesh.renderItems: couldn't find render method for ",
				print self.__uriForHomeToolAndUniqueId(toolName, item.uniqueId)
			return rendered
		finally: self._release()

__desc__ = _("This tool allows to configure stray link resolution")

PyDS.Tool.registerTool(MeshTool, 'mesh', _('Mesh'), __desc__, 905)
PyDS.StructuredText.registerTransform(StrayLinkTransform)

