"""

Python Desktop Server - Mirror Tool

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: MirrorTool.py,v 1.20 2004/02/04 18:06:19 gb Exp $

import sys
import time
import string
import traceback

import PyDS.Tool

from PyDS.PersistentQueue import PersistentQueue

import PyDS.MirrorRegistry
import PyDS.PyCSSearchMirror

class MirrorTool(PyDS.Tool.StandardTool):

	def _createdb(klass, db):
		db.getas("mirrors[tag:S,active:I]")

	_createdb = classmethod(_createdb)

	def _initdb(self):
		try:
			self._acquire()
			self.mirrors = self.db.view('mirrors').ordered(1)
			self._commit()
			self.queue = PersistentQueue(self.name, 'mirror')
		finally: self._release()
	
	def _initopts(self):
		PyDS.MirrorRegistry.instantiateMirrors(self)

	def _initactive(self):
		timer = self.getToolByNamespace('timer')
		navigation = self.getToolByNamespace('navigation')
		timer._addHook('select', 'mirror', _('resync mirrors'), self._resyncTimerHook)
		navigation._addHook('select', 'mirror',
			_('Link to the search page'),
			self._searchNavHook,
			self._searchNavHookActive)

	def _initthread(self, server):
		prefs = self.getPreferencesTool()
		while server.isRunning():
			time.sleep(10)
			if self.queue.length() and prefs.online():
				while self.queue.length():
					(cmd, key, data) = self.queue.shift()
					jobf = getattr(self, cmd)
					jobf(key, data)
				self.logVerbose(_('work done, going to sleep again'))

	def _shutdown(self):
		PyDS.MirrorRegistry.delMirrors()

	def _status(self):
		html = PyDS.Tool.StandardTool._status(self)
		html += '<ul>'
		html += _('<li>There are %d postings in the index queue</li>') % self.queue.length()
		html += self._threadstatus()
		html += '</ul>'
		return html

	def _resyncTimerHook(self, qual, item=None):
		if qual == 'title':
			return _('resync mirrors')
		elif qual == 'select':
			html = '<p><select name="item">'
			for mirror in PyDS.MirrorRegistry.getMirrors():
				if self.isMirrorActive(mirror.tag) and mirror.canWork():
					html += '<option value="%s">%s' % (
						mirror.tag,
						mirror.title
					)
			html += '</select>'
			return html
		elif qual == 'item':
			return _('resyncing mirror "%s"') % item
		elif qual == 'run':
			self.resync(mirror=item)
		else:
			raise ValueError('%s is an unknown qualifier' % qual)

	def _searchNavHook(self, qual, value, preset=None):
		if qual == 'html':
			return ''
		elif qual == 'url':
			upstream = self.getToolByNamespace('upstream')
			url = upstream.get('prefs', 'urlsearch')
			return url
		else:
			raise ValueError('%s is not a valid qualifier' % qual)
	
	def _searchNavHookActive(self):
		mirror = PyDS.MirrorRegistry.getMirror('pycsSearch')
		return mirror and self.isMirrorActive(mirror.tag) and mirror.canWork()
	
	def _addItem(self, key, data):
		(tool, uniqueId) = data.split(':')
		tool = self.getToolByNamespace(tool)
		rssHash = tool.getRSSItem(uniqueId, fullrss=1)
		for mirror in PyDS.MirrorRegistry.getMirrors():
			if self.isMirrorActive(mirror.tag) and mirror.canWork():
				mirror.addItem(tool, uniqueId, rssHash)

	def _updateItem(self, key, data):
		(tool, uniqueId) = data.split(':')
		tool = self.getToolByNamespace(tool)
		rssHash = tool.getRSSItem(uniqueId, fullrss=1)
		for mirror in PyDS.MirrorRegistry.getMirrors():
			if self.isMirrorActive(mirror.tag) and mirror.canWork():
				mirror.updateItem(tool, uniqueId, rssHash)

	def _deleteItem(self, key, data):
		(tool, uniqueId) = data.split(':')
		tool = self.getToolByNamespace(tool)
		for mirror in PyDS.MirrorRegistry.getMirrors():
			if self.isMirrorActive(mirror.tag) and mirror.canWork():
				mirror.deleteItem(tool, uniqueId)
	
	def _syncItem(self, key, data):
		(tool, uniqueId) = data.split(':')
		tool = self.getToolByNamespace(tool)
		rssHash = tool.getRSSItem(uniqueId, fullrss=1)
		for mirror in PyDS.MirrorRegistry.getMirrors():
			if self.isMirrorActive(mirror.tag) and mirror.canWork():
				mirror.syncItem(tool, uniqueId, rssHash)
	
	def _resyncItems(self, key, data):
		if data:
			mirror = PyDS.MirrorRegistry.getMirror(data)
			if self.isMirrorActive(mirror.tag) and mirror.canWork():
				for tool in PyDS.Tool.getTools():
					if hasattr(tool, '_resync'):
						mirror._resync(tool)
						self.logVerbose(_('delete/update/add for mirror "%s" on tool "%s": %d/%d/%d') % (mirror.tag, tool.name, todel, toupdate, toadd))
		else:
			for mirror in PyDS.MirrorRegistry.getMirrors():
				if self.isMirrorActive(mirror.tag) and mirror.canWork():
					for tool in PyDS.Tool.getTools():
						if hasattr(tool, '_resync'):
							(todel, toupdate, toadd) = mirror._resync(tool)
							self.logVerbose(_('delete/update/add for mirror "%s" on tool "%s": %d/%d/%d') % (mirror.tag, tool.name, todel, toupdate, toadd))

	def index_html(self, req):
		try:
			self._acquire()
			html = '<form action="%s" method="POST">' % self.getUrl('activate_redir')
			html += '<table width="100%" border=0>'
			html += '<tr><th>%s</th><th>%s</th></tr>' % (
				_('Active'),
				_('Mirror')
			)
			for mirror in PyDS.MirrorRegistry.getMirrors():
				html += '<tr class="whitebox">'
				if mirror.canWork():
					if self.isMirrorActive(mirror.tag):
						html += '<td><input type="checkbox" name="active" value="%s" checked></td>' % mirror.tag
					else:
						html += '<td><input type="checkbox" name="active" value="%s"></td>' % mirror.tag
				else:
					html += '<td>&nbsp;</td>'
				html += '<td>%s (%d %s)</td>' % (
					mirror.title,
					mirror.getItemCount(),
					_('elements')
				)
				html += '</tr>'
			html += '<tr><td align="right" colspan=2>'
			html += '<input type="submit" name="acitvate" value="%s">' % _('Save')
			html += '</td></tr>'
			html += '</table>'
			html += '</form>'
			html += '<p><a href="%s">%s</a>' % (
				req.getUrl('resync_redir'),
				_('Click here to resync all active mirrors with your database content.')
			)
			req.setLocalValue('title', _('List of registered mirrors'))
			return req.renderPage('BaseTemplate', body=html)
		finally: self._release()
	
	def activate_redir(self, req):
		mhash = {}
		for mirror in PyDS.MirrorRegistry.getMirrors():
			mhash[mirror.tag] = 0
		for tag in req.getAllValues('active'):
			mhash[tag] = 1
		for tag in mhash.keys():
			if mhash[tag] and not self.isMirrorActive(tag):
				self.activateMirror(tag)
			elif not mhash[tag] and self.isMirrorActive(tag):
				self.deactivateMirror(tag)
		return req.getUrl()

	def resync_redir(self, req):
		self.resync()
		return req.getUrl()

	# this method checks wether a mirror is active
	def isMirrorActive(self, tag):
		(idx, found) = self.mirrors.locate({'tag':tag})
		if found:
			return self.mirrors[idx].active
		else:
			return 0
	
	# this method activates a mirror - posts will be replicated from then on
	def activateMirror(self, tag):
		try:
			self._acquire()
			(idx, found) = self.mirrors.locate({'tag':tag})
			if found: self.mirrors[idx].active = 1
			else: self.mirrors.append({'tag':tag, 'active':1})
			self._commit()
		finally: self._release()
	
	# this method deactivates a mirror - posts won't be replicated from then on
	def deactivateMirror(self, tag):
		try:
			self._acquire()
			(idx, found) = self.mirrors.locate({'tag':tag})
			if found: self.mirrors[idx].active = 0
			else: self.mirrors.append({'tag':tag, 'active':0})
			self._commit()
		finally: self._release()
	
	# this method resyncs all mirrors with all tools providing this feature
	def resync(self, mirror=''):
		self.queue.append('_resyncItems', 'resync:%s' % mirror, mirror)
	
	# this method replicates a new post to all mirrors. It will fetch the
	# item with getRSSItem from the given tool.
	# This is called after the item creation, the hash contains the content.
	def addItem(self, tool, uniqueId):
		if isinstance(tool, PyDS.Tool.StandardTool):
			tool = tool.name
		self.queue.append('_addItem', 'add:%s:%s' % (tool, uniqueId), '%s:%s' % (tool, uniqueId))
		
	# this method replicates an updated post to all active mirrors.
	# This is called after the actual update, so getRSSItem will fetch
	# the new values. If you need old values, keep them around in your mirror.
	def updateItem(self, tool, uniqueId):
		if isinstance(tool, PyDS.Tool.StandardTool):
			tool = tool.name
		self.queue.append('_updateItem', 'update:%s:%s' % (tool, uniqueId), '%s:%s' % (tool, uniqueId))
		
	# this method replicates a post to all active mirrors, doing
	# either the add or update automatically. This is called after the
	# actual update, so getRSSItem will fetch the new values. If you need
	# old values, keep them around in your mirror.
	def syncItem(self, tool, uniqueId):
		if isinstance(tool, PyDS.Tool.StandardTool):
			tool = tool.name
		self.queue.append('_syncItem', 'sync:%s:%s' % (tool, uniqueId), '%s:%s' % (tool, uniqueId))
		
	# this method replicates a deleted post to all active mirrors. This is
	# called after the actual deletion takes place, so you can't fetch the
	# old content! If you need stuff, keep it in the mirror database.
	def deleteItem(self, tool, uniqueId):
		if isinstance(tool, PyDS.Tool.StandardTool):
			tool = tool.name
		self.queue.append('_deleteItem', 'delete:%s:%s' % (tool, uniqueId), '%s:%s' % (tool, uniqueId))
	
	# this method returns a searchbox for your layout. It's only returned if your search actually contains
	# any searchable items (your community server has returned hassearch with a true value!)
	def searchbox(self):
		upstream = self.getToolByNamespace('upstream')
		macros = self.getToolByNamespace('macros')
		mirror = PyDS.MirrorRegistry.getMirror('pycsSearch')
		if upstream.get('prefs','hassearch') and mirror and self.isMirrorActive(mirror.tag) and mirror.canWork():
			html = '<p>'
			html += '<form class="whiteboxsmall" action="%s">' % upstream.get('prefs','urlsearch')
			html += '<input type="hidden" name="u" value="%s">' % upstream.get('prefs', 'usernum')
			html += '<input name="q" size=10>'
			html += '<input type="image" src="%s" alt="%s">' % (
				macros.absoluteUrl('/images/search.gif'),
				_('search')
			)
			html += '</form>'
			return html
		else:
			return ''

		
__desc__ = _("This tool controls all registered item mirrors.")

PyDS.Tool.registerTool(MirrorTool, 'mirror', _('Mirror'), __desc__, 902)

