"""

Python Desktop Server - Timer to periodically trigger actions

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: TimerTool.py,v 1.23 2004/02/09 16:43:30 gb Exp $

import sys
import time
import traceback

import PyDS.Tool

class TimerTool(PyDS.Tool.StandardTool):

	def _createdb(klass, db):
		db.getas("timers[id:I,hook:S,item:S,interval:I,last:I]")
		db.getas("prefs[maxid:I]")

	_createdb = classmethod(_createdb)

	def _initdb(self):
		try:
			self._acquire()
			self.timers = self.db.view('timers')
			self.prefs = self.db.view('prefs')
			if len(self.prefs) == 0:
				self.prefs.append({'maxid':0})
			self._commit()
			self.importantTool = 1
		finally: self._release()
	
	def _initopts(self):
		self.hasAPI = 0
	
	def _backup(self):
		l = self._backupTable('prefs')
		l.extend(self._backupTable('timers'))
		return l
	
	def _readyTimers(self):
		try:
			self._acquire()
			now = time.time()
			torun = []
			for timer in self.timers:
				if timer.last + timer.interval*60 < now:
					torun.append((timer.hook, timer.item))
					while timer.last + timer.interval*60 < now:
						timer.last += timer.interval*60
			if len(torun):
				self._commit()
			return torun
		finally: self._release()
	
	def _initthread(self, server):
		while server.isRunning():
			time.sleep(10)
			torun = self._readyTimers()
			if len(torun):
				for (hookname, item) in torun:
					try:
						hook = self._getHook('select', hookname)
						self.log(hook('item', item))
						hook('run', item)
					except:
						(e, d, tb) = sys.exc_info()
						msg = _('Exception "%s: %s" in timer hook "%s"') % (e, d, hookname)
						for row in traceback.extract_tb(tb):
							(file, line, function, source) = row
							msg += '<br>%s[%s] in %s' % (file, line, function)
						self.logError(msg)
				self.logVerbose('work done, going to sleep again')
	
	def _status(self):
		try:
			self._acquire()
			html = PyDS.Tool.StandardTool._status(self)
			html += '<ul>'
			html += _('<li>There are %d timer defined</li>') % len(self.timers)
			html += self._threadstatus()
			html += '</ul>'
			return html
		finally: self._release()
	
	def getPreferencePanels(self):
		return [
			['timer_panel', _('Timer'),
				_('Here you can define your periodic activities.')]
		]
			
	def timer_panel(self, req):
		macros = self.getToolByNamespace('macros')
		html = _('<p>All defined timers are listed here.')
		if len(self.timers):
			html += '<form action="%s" method="POST">' % self.getUrl('timer_redir')
			html += '<table border=0 width="100%">'
			html += '<tr>'
			html += '<th>%s</th><th align="left">%s</th>' % (_('Delete'), _('Timer'))
			html += '</tr>'
			for timer in self.timers:
				html += '<tr>'
				html += '<td>'
				html += '<input type=checkbox name="timer" value="%s">' % timer.id
				html += '</td>'
				hook = self._getHook('select', timer.hook)
				try:
					html += '<td>%s <span class="smallfont">(%s, %s %s)</span></td>' % (hook('item', timer.item), _('every %s minutes') % timer.interval, _('next run at'), macros.Today(now=timer.last+timer.interval*60,withtime=1))
				except:
					html += '<td>[%s] <span class="smallfont">(%s)</span></td>' % (timer.item, _('every %s minutes') % timer.interval)
				html += '</tr>'
			html += '</table>'
			html += '<input type=submit name="DeleteSubmit" value="%s">' % _('delete')
			html += ' <input type=submit name="StartSubmit" value="%s">' % _('run now')
			html += '</form>'
		html += '<p><form action="%s" method="POST">' % (
			self.getUrl('addtimer_redir')
		)
		html += '<select name="timer">'
		for hook in self._getHooks('select'):
			if hook.isActive():
				html += '<option value="%s">%s' % (hook.getId(), hook.title)
		html += '</select>'
		html += '<input type=submit name="AddTimerSubmit" value="%s">' % _('Add')
		html += '</form>'
		return html
	
	def addtimer_panel(self, req):
		hookname = req.getFirstValue('timer')
		hook = self._getHook('select', hookname)
		html = '<p>%s %s' % (_('Add a new timer:'), hook.title)
		html += '<form action="%s" method="POST">' % self.getUrl('doaddtimer_redir')
		html += '<input type=hidden name="timer" value="%s">' % hookname
		html += hook('select')
		html += _('<p>Run this timer every %s minutes, ') % (
			'<input name="interval" size=5>',
		)
		(hh,mm) = time.localtime(time.time()+15*60)[3:5]
		html += _('starting at %s') % (
			'<input name="time" size=5 value="%02d:%02d">' % (hh,mm),
		)
		html += '<p><input type=submit name="DoAddTimerSubmit" value="%s">' % _('Add')
		html += '</form>'
		return html
	
	def timer_redir(self, req):
		try:
			self._acquire()
			liste = req.getAllValues('timer')
			delete = req.getFirstValue('DeleteSubmit')
			start = req.getFirstValue('StartSubmit')
			if liste:
				if delete:
					todel = []
					for id in liste:
						if type(id) != type(1):
							id = int(id)
						(idx, found) = self.timers.locate({'id':id})
						if found:
							todel.append(idx)
					todel.sort()
					todel.reverse()
					for idx in todel:
						self.timers.delete(idx)
					self._commit()
				elif start:
					now = int(time.time())
					for id in liste:
						if type(id) != type(1):
							id = int(id)
						(idx, found) = self.timers.locate({'id':id})
						if found:
							self.timers[idx].last = now - (self.timers[idx].interval + 1) * 60
					self._commit()
			return req.getUrl(
				'panel_html',
				tool='preferences',
				panel='timer_panel',
				fortool='timer'
			)
		finally: self._release()
	
	def addtimer_redir(self, req):
		return req.getUrl(
			'panel_html',
			tool='preferences',
			panel='addtimer_panel',
			fortool='timer',
			timer=req.getFirstValue('timer')
		)
	
	def doaddtimer_redir(self, req):
		try:
			self._acquire()
			hookname = req.getFirstValue('timer')
			hook = self._getHook('select', hookname)
			item = req.getFirstValue('item')
			interval = int(req.getFirstValue('interval'))
			firstrun = req.getFirstValue('time').split(':')
			now = time.time()
			t = list(time.localtime(now))
			t[3] = int(firstrun[0])
			t[4] = int(firstrun[1])
			tt = time.mktime(t)
			if tt < now:
				tt += 24 * 60 * 60
			tt -= interval * 60
			res = self.timers.select({'hook':hookname, 'item':item})
			if len(res) == 0:
				self.prefs[0].maxid += 1
				self.timers.append({
					'id':self.prefs[0].maxid,
					'hook':hookname,
					'item':item,
					'interval':interval,
					'last':int(tt)
				})
			self._commit()
			return req.getUrl(
				'panel_html',
				tool='preferences',
				panel='timer_panel',
				fortool='timer'
			)
		finally: self._release()

__desc__ = _("This tool manages timers for periodic execution.")

PyDS.Tool.registerTool(TimerTool, 'timer', _('Timer'), __desc__, 1400)

