"""

Python Desktop Server - manage Users for remote access

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: UserTool.py,v 1.14 2004/02/09 16:43:30 gb Exp $

import sys
import md5
import time
import poplib

try:
	import smbauth
	smbavail = 1
except:
	smbavail = 0

import PyDS.Tool

class UserTool(PyDS.Tool.StandardTool):

	def _createdb(klass, db):
		db.getas("users[id:S,admin:I,method:S,parms:S]")
		db.getas("adminlevels[tool:S,needadmin:I]")

	_createdb = classmethod(_createdb)

	def _initdb(self):
		try:
			self._acquire()
			self.users = self.db.view('users').ordered(1)
			self.adminlevels = self.db.view('adminlevels').ordered(1)
			self._methods = {
				'local': _('local user with password'),
				'pop3': _('remote user with pop3 authorization'),
			}
			if smbavail:
				self._methods['smb'] = _('remote user with NT authorization')
			self.importantTool = 1
		finally: self._release()
	
	def _initopts(self):
		try:
			self._acquire()
			for row in self.adminlevels:
				PyDS.Tool.setNeedAdmin(row.tool, row.needadmin)
		finally: self._release()
	
	def _backup(self):
		l = self._backupTable('adminlevels')
		return l

	def _setAdminNeeded(self, tool, flag):
		try:
			self._acquire()
			(idx, found) = self.adminlevels.locate({'tool':tool.name})
			if found:
				self.adminlevels[idx].needadmin = flag
			else:
				self.adminlevels.append({
					'tool':tool.name,
					'needadmin':flag
				})
			self._commit()
			tool.needAdmin = flag
		finally: self._release()
	
	def _status(self):
		try:
			self._acquire()
			html = PyDS.Tool.StandardTool._status(self)
			html += '<ul>'
			html += _('<li>There are %d users defined</li>') % len(self.users)
			html += '</ul>'
			return html
		finally: self._release()
	
	def nugget(self, url=None):
		if self.authenticated_user:
			html = _('You are logged in as <b>%s</b>') % self.authenticated_user
			url = self.getUrl(
				'panel_html',
				tool='preferences',
				panel='user_panel',
				fortool='user'
			)
			return '<a href="%s">%s</a>: %s' % (url, self.title, html)
		elif url:
			return '<a href="%s">%s</a>: %s' % (url, self.title, self.desc)
		else:
			return ''
	
	def checkUser(self, user, password, needadmin=0):
		try:
			self._acquire()
			(idx, found) = self.users.locate({'id': user})
			if found:
				usr = self.users[idx]
				if usr.method == 'local':
					if usr.parms != md5.md5(password).hexdigest():
						self.log(_('invalid password given for local user %s'), user)
						return 0
				elif usr.method == 'pop3':
					try:
						srv = poplib.POP3(usr.parms)
						srv.user(user)
						srv.pass_(password)
						srv.quit()
					except:
						(e, d, tb) = sys.exc_info()
						self.log(
							_('login to pop3 failed with exception "%s: %s" for user %s'),
							e,
							d,
							user
						)
						return 0
				elif usr.method == 'smb':
					(pdc, bdc, domain) = usr.parms.split(':')
					res = smbauth.smbauth_validate(user, password, pdc, bdc, domain)
					if res:
						msg = smbauth.smbauth_err2str(res)
						self.log(_('login to smb gave "%s" for user %s'), msg, user)
						return 0
				if needadmin and not(usr.admin):
					self.log(_('admin needed but user %s is no admin'), user)
					return 0
				return 1
			else:
				self.log(_('user %s could not be found'), user)
				return 0
		finally: self._release()
	
	def getPreferencePanels(self):
		return [
			['user_panel', _('User'),
				_('Configure users with remote access rights.')],
			['tool_panel', _('Adminlevel'),
				_('Configure admin level for tools')]
		]
			
	def user_panel(self, req):
		html = _('<p>All defined users in this system are listed here.')
		if len(self.users):
			html += '<form action="%s" method="POST">' % self.getUrl('user_redir')
			html += '<table border=0 width="100%">'
			html += '<tr>'
			html += '<th>%s</th><th align="left">%s</th><th>%s</th>' % (
				_('Delete'),
				_('User'),
				_('Info')
			)
			html += '</tr>'
			for user in self.users.sort(self.users.id):
				html += '<tr>'
				html += '<td>'
				html += '<input type=checkbox name="user" value="%s">' % user.id
				html += '</td><td>'
				html += '<a href="%s">%s</a>' % (
					self.getUrl(
						'panel_html',
						tool='preferences',
						panel='edit_panel',
						fortool='user',
						user=user.id
					),
					user.id
				)
				html += '</td><td>'
				if user.admin:
					html += _('<b>admin</b>, ')
				html += self._methods.get(user.method, user.method)
				html += '</td></tr>'
			html += '</table>'
			html += '<input type=submit name="UserSubmit" value="%s">' % _('delete')
			html += '</form>'
		html += '<p><form action="%s" method="POST">' % (
			self.getUrl('panel_html', tool='preferences')
		)
		html += '<input type="hidden" name="panel" value="edit_panel">'
		html += '<input type="hidden" name="fortool" value="user">'
		html += '<select name="method">'
		for method in self._methods.keys():
				html += '<option value="%s">%s' % (method, self._methods[method])
		html += '</select>'
		html += ' <input type=submit name="AddUserSubmit" value="%s">' % _('Add')
		html += '</form>'
		return html
	
	def edit_panel(self, req):
		method = req.getFirstValue('method')
		user = req.getFirstValue('user')
		admin = 0
		parms = ''
		if user:
			try:
				self._acquire()
				(idx, found) = self.users.locate({'id':user})
				if found:
					usr = self.users[idx]
					admin = usr.admin
					method = usr.method
					parms = usr.parms.split(':')
			finally: self._release()
		html = '<p>%s %s' % (_('Edit user:'), self._methods[method])
		html += '<form action="%s" method="POST">' % self.getUrl('edit_redir')
		html += '<input type=hidden name="method" value="%s">' % method
		if user:
			html += '<input type=hidden name="user" value="%s">' % user
		html += '<table width="100%" border=0>'
		if user:
			html += '<tr><th>%s</th><td>%s</td></tr>' % (
				_('user'), user
			)
		else:
			html += '<tr><th>%s</th><td><input name="user"></td></tr>' % (
				_('user')
			)
		html += '<tr>'
		html += '<th>%s</th><td>' % _('Admin')
		html += '<select name="admin">'
		if admin:
			html += '<option value="0">%s' % _('no')
			html += '<option value="1" selected>%s' % _('yes')
		else:
			html += '<option value="0" selected>%s' % _('no')
			html += '<option value="1">%s' % _('yes')
		html += '</select>'
		html += '</td>'
		html += '</tr>'
		if method == 'local':
			html += '<tr><th>%s</th><td><input type="password" name="password"></td></tr>' % (
				_('password')
			)
			html += '<tr><th>%s</th><td><input type="password" name="repeat"></td></tr>' % (
				_('repeat')
			)
		elif method == 'pop3':
			server = ''
			if parms:
				server = parms[0]
			html += '<tr><th>%s</th><td><input name="server" value="%s"></td></tr>' % (
				_('server'),
				server
			)
		elif method == 'smb':
			pdc = ''
			bdc = ''
			domain = ''
			if parms:
				pdc, bdc, domain = parms
			html += '<tr><th>%s</th><td><input name="pdc" value="%s"></td></tr>' % (
				_('PDC'),
				pdc
			)
			html += '<tr><th>%s</th><td><input name="bdc" value="%s"></td></tr>' % (
				_('BDC'),
				bdc
			)
			html += '<tr><th>%s</th><td><input name="domain" value="%s"></td></tr>' % (
				_('Domain'),
				domain
			)
		html += '</table>'
		html += '<p><input type=submit name="DoAddUserSubmit" value="%s">' % _('Save')
		html += '</form>'
		return html
	
	def edit_redir(self, req):
		try:
			self._acquire()
			user = req.getFirstValue('user')
			admin = int(req.getFirstValue('admin'))
			method = req.getFirstValue('method')
			parms = ''
			if method == 'local':
				password = req.getFirstValue('password')
				repeat = req.getFirstValue('repeat')
				if password == repeat:
					if password:
						parms = md5.md5(password).hexdigest()
				else:
					return req.getUrl(
						'panel_html',
						tool='preferences',
						panel='edit_panel',
						fortool='user',
						user=user,
						method=method,
						errmsg=_('the repeated password was not identical')
					)
			elif method == 'pop3':
				parms = req.getFirstValue('server')
			elif method == 'smb':
				parms = '%s:%s:%s' % (
					req.getFirstValue('pdc'),
					req.getFirstValue('bdc'),
					req.getFirstValue('domain')
				)
			if user:
				(idx, found) = self.users.locate({'id':user})
				if found:
					usr = self.users[idx]
					if parms:
						usr.parms = parms
					usr.admin = admin
				else:
					if not(parms):
						return req.getUrl(
							'panel_html',
							tool='preferences',
							panel='edit_panel',
							fortool='user',
							user=user,
							method=method,
							errmsg=_('you must fill in the parameters')
						)
					self.users.append({
						'id': user,
						'method': method,
						'parms': parms,
						'admin': admin
					})
				self._commit()
				return req.getUrl(
					'panel_html',
					tool='preferences',
					panel='user_panel',
					fortool='user'
				)
			else:
				return req.getUrl(
					'panel_html',
					tool='preferences',
					panel='edit_panel',
					fortool='user',
					method=method,
					errmsg=_('you must give a user name')
				)
		finally: self._release()
	
	def user_redir(self, req):
		try:
			self._acquire()
			liste = req.getAllValues('user')
			if liste:
				todel = []
				for id in liste:
					(idx, found) = self.users.locate({'id':id})
					if found:
						todel.append(idx)
				todel.sort()
				todel.reverse()
				for idx in todel:
					self.users.delete(idx)
				self._commit()
			return req.getUrl(
				'panel_html',
				tool='preferences',
				panel='user_panel',
				fortool='user'
			)
		finally: self._release()
	
	def tool_panel(self, req):
		macros = self.getToolByNamespace('macros')
		html = _('<p>Here you can configure which of your content tools need admin level to post content or use forms. If you disable the need for admin level, this toll can be fully used by users not defined as admins. Otherwise those users can only view content, but not submit forms.')
		html += '<form action="%s" method="POST">' % self.getUrl('tool_redir')
		html += '<table border="0">'
		html += '<tr><th align="left">%s</th><th align="left">%s</th><th align="left">%s</th></tr>' % (
			_('Admin<br>required?'),
			_('Tool'),
			_('Description')
		)
		for tool in PyDS.Tool.getTools():
			if isinstance(tool, PyDS.Tool.StandardTool):
				html += '<tr>'
				html += '<td>'
				html += macros.checkBox('tool.%s' % tool.name, tool.needAdmin)
				html += '</td>'
				html += '<td class="whiteboxsmall">%s</td>' % tool.title
				html += '<td class="whiteboxsmall">%s</td>' % tool.desc
				html += '</tr>'
		html += '<tr><td align="right" colspan=3>'
		html += '<input type=submit name=submit value="%s">' % _('Save')
		html += '</td></tr>'
		html += '</table>'
		html += '</form>'
		return html
	
	def tool_redir(self, req):
		for tool in PyDS.Tool.getTools():
			if isinstance(tool, PyDS.Tool.StandardTool):
				val = req.getFirstValue('tool.%s' % tool.name)
				if val != '':
					self._setAdminNeeded(tool, int(val))
		return req.getUrl(
			'panel_html',
			tool='preferences',
			panel='tool_panel',
			fortool='user'
		)

__desc__ = _("This tool manages users with remote access rights.")

PyDS.Tool.registerTool(UserTool, 'user', _('User'), __desc__, 800)

